import copy

import numpy as np
import scipy.stats as sps
import matplotlib.pyplot as plt
from matplotlib.widgets import CheckButtons,Button



def default_tool_tips(result_object, learning_curves, include_run_info=False):

	tool_tips = {}
	id2conf = result_object.get_id2config_mapping()
	
	for id in learning_curves.keys():
		
		config = id2conf[id]['config']
		config_info = id2conf[id]['config_info']

		all_runs =  result_object.get_runs_by_id(id)
		if len(all_runs) == 0:
			continue
		longest_run = all_runs[-1]
	
		while longest_run.loss is None:
			all_runs.pop()
			if len(all_runs) == 0: break
			longest_run = all_runs[-1]
			
		if len(all_runs) == 0: continue
		
		s = ['id: %s'%str(id), 'duration (sec): %f'%((longest_run['time_stamps']['finished'] - longest_run['time_stamps']['started']))]

		if not longest_run.loss is None:
			s += [str(k) + "=" +str(v) for k,v in sorted(id2conf[id]['config'].items()) ]
			try:
				s += [str(k) + "=" +str(v) for k,v in sorted(id2conf[id]['config_info'].items()) ]
			except:
				pass
			
			s += ['losses: {}'.format([r.loss for r in all_runs])]
			if include_run_info:
				s += ['longest run info: {}'.format(longest_run.info)]
			
		tool_tips[id] = "\n".join(s)
	return(tool_tips)

def concurrent_runs_over_time(runs, num_points = 512, show=False):
	
	data = np.array([(r.time_stamps['started'], r.time_stamps['finished']) for r in runs])
	ts = np.linspace(data.min(), data.max(), 512)
	n_workers = np.array([ ((data[:,0] <= t)*(data[:,1]>t)).sum() for t in ts])

	fig, ax = plt.subplots()
	ax.plot(ts, n_workers)
	ax.set_xlabel('time [s]')
	ax.set_ylabel('number of concurent runs')
	
	if show:
		plt.show()
	return(fig, ax)


def finished_runs_over_time(runs, show=False):
	budgets = set([r.budget for r in runs])
	
	times = {}
	for b in budgets:
		times[b] = [0]
	
	for r in runs:
		times[r.budget].append(r.time_stamps['finished'])
	
	for b in budgets:
		times[b].sort()
	
	
	fig, ax = plt.subplots()

	for b in budgets:
		ax.plot(times[b], np.arange(len(times[b])), label='b = %f'%b)


	ax.set_xlabel('time [s]')
	ax.set_ylabel('number of finished runs')
	ax.legend()
	
	if show:
		plt.show()
	return(fig,ax)

def performance_histogram_model_vs_random(runs, id2conf, show=False):
	model_based_runs = list(filter(lambda r: id2conf[r.config_id]['config_info']['model_based_pick'], runs))
	random_runs = list(filter(lambda r: not id2conf[r.config_id]['config_info']['model_based_pick'], runs))

	budgets = list(set([r.budget for r in runs]))
	budgets.sort()

	losses = {}
	for b in budgets:
		losses[b] = {'model_based': [], 'random': []}

	for r in model_based_runs:
		if r.loss is None or not np.isfinite(r.loss):
			continue
		losses[r.budget]['model_based'].append(r.loss)

	for r in random_runs:
		if r.loss is None or not np.isfinite(r.loss):
			continue
		losses[r.budget]['random'].append(r.loss)

	fig, axarr = plt.subplots(len(budgets), 2, sharey='row', sharex='row')
	plt.suptitle('Loss of model based configurations (left) vs. random configuration (right)')

	for i,b in enumerate(budgets):
		mbax, rax = axarr[i]
		mbax.hist(losses[b]['model_based'], label='b = %f \n n = %i'%(b,len(losses[b]['model_based'])))
		mbax.set_ylabel('frequency')
		mbax.legend()
		
		
		
		rax.hist(losses[b]['random'],label='b = %f \n n = %i'%(b,len(losses[b]['random'])))
		rax.legend()
		
		
		if i == len(budgets)-1:
			mbax.set_xlabel('loss')
			rax.set_xlabel('loss')
	if show:		
		plt.show()
		
	return(fig, axarr)


def correlation_across_budgets(results_object, show=False):
	
	runs = results_object.get_all_runs()
	id2conf = results_object.get_id2config_mapping()

	budgets = list(set([r.budget for r in runs]))
	budgets.sort()

	import itertools

	loss_pairs = {}
	for b in budgets[:-1]:
		loss_pairs[b] = {}

	for b1,b2 in itertools.combinations(budgets, 2):
		loss_pairs[b1][b2]= []

	for cid in id2conf.keys():
		runs = results_object.get_runs_by_id(cid)
		if len(runs) < 2: continue
		
		for r1,r2 in itertools.combinations(runs,2):
			if r1.loss is None or r2.loss is None: continue
			if not np.isfinite(r1.loss) or not np.isfinite(r2.loss): continue
			loss_pairs[float(r1.budget)][float(r2.budget)].append((r1.loss, r2.loss))
		
		

	rhos = np.eye(len(budgets)-1)
	rhos.fill(np.nan)

	ps = np.eye(len(budgets)-1)
	ps.fill(np.nan)

	for i in range(len(budgets)-1):
		for j in range(i+1,len(budgets)):
			spr = sps.spearmanr(loss_pairs[budgets[i]][budgets[j]])
			rhos[i][j-1] = spr.correlation
			ps[i][j-1] = spr.pvalue


	fig, ax = plt.subplots()

	cax = ax.matshow(rhos, vmin=-1, vmax=1)
	fig.colorbar(cax)


	ax.set_yticks( range(len(budgets)-1))
	ax.set_yticklabels(budgets[:-1],)

	ax.set_xticks( range(len(budgets)-1))
	ax.set_xticklabels(budgets[1:],)
	
	ax.set_title('Rank correlation of the loss across the budgets')

	for i in range(len(budgets)-1):
		for j in range(i+1,len(budgets)):
			plt.text(j-1,i, r'$\rho_{spearman}= %f$'%rhos[i][j-1] + '\n' + r'$p = %f$'%ps[i][j-1] + '\n' + r'$n = %i$'%len(loss_pairs[budgets[i]][budgets[j]]),
						horizontalalignment='center', verticalalignment='center')

	if show:
		plt.show()
	return(fig,ax)



def losses_over_time(runs, get_loss_from_run_fn = lambda r: r.loss, cmap = plt.get_cmap("tab10"), show=False):

	budgets = set([r.budget for r in runs])

	data = {}
	for b in budgets:
		data[b] = []

	for r in runs:
		if r.loss is None:
			continue
		b = r.budget
		t = r.time_stamps['finished']
		l = get_loss_from_run_fn(r)
		data[b].append((t,l))

	for b in budgets:
		data[b].sort()


	fig, ax = plt.subplots()

	for i, b in enumerate(budgets):
		data[b] = np.array(data[b])
		ax.scatter(data[b][:,0], data[b][:,1], color=cmap(i), label='b=%f'%b)
		
		ax.step(data[b][:,0], np.minimum.accumulate(data[b][:,1]), where='post')

	ax.set_title('Losses for different budgets over time')
	ax.set_xlabel('wall clock time [s]')
	ax.set_ylabel('loss')
	ax.legend()
	if show:
		plt.show()
	return(fig,ax)





def interactive_HBS_plot(learning_curves, tool_tip_strings=None,log_y=False, log_x=False, reset_times=False, color_map='Set3', colors_floats=None, title='', show=True):

	times, losses, config_ids, = [], [], []

	for k,v in learning_curves.items():
		for l in v:
			if len(l) == 0: continue
			tmp = list(zip(*l))
			try:
				times.append(tmp[0])
				losses.append(tmp[1])
				config_ids.append(k)
			except:
				import pdb; pdb.set_trace()



	num_curves = len(times)
	HB_iterations = [id[0] for id in config_ids]
	
	num_iterations = len(set(HB_iterations))
	
	cmap = plt.get_cmap(color_map)
	

	
	if reset_times:
		times = [np.array(ts) - ts[0] for ts in times]
	
	
	if colors_floats is None:
		color_floats = []
		for i in range(num_curves):
			seed = 100*np.abs(config_ids[i][0]) + 10*config_ids[i][1] + config_ids[i][2]
			np.random.seed(seed)
			color_floats.append(np.random.rand())

	fig, ax = plt.subplots()
	
	lines = [[] for i in range(num_iterations)]

	iteration_labels  = list(range(num_iterations))
	if HB_iterations[-1] == -1:
		iteration_labels[-1] = 'warmstart data'
	
	

	all_lines = []
	
	for i in range(num_curves):
		l, = ax.plot(times[i], losses[i], color=cmap(color_floats[i]), marker='o', gid=i, picker=True)
		lines[HB_iterations[i]].append(l)
		all_lines.append(l)

	if log_y:
		plt.yscale('log')

	ax.set_title(title)

	hover_annotation = ax.annotate("BLABLA", xy=(0,0), xytext=(20,20),textcoords="offset points",
					bbox=dict(boxstyle="round", fc="w"),
					arrowprops=dict(arrowstyle="->"))
	hover_annotation.set_visible(False)


	permanent_annotations = {}


	plt.subplots_adjust(left=0.2)
	rax = plt.axes([0.05, 0.1, 0.1, 0.8])

	
	axnone = plt.axes([0.05, 0, 0.05, 0.1])
	axall = plt.axes([0.1, 0, 0.05  , 0.1])
		
	check = CheckButtons(rax, iteration_labels, [True for i in range(num_iterations)])
	
	none_button = Button(axnone, 'None')
	all_button = Button(axall, 'All')
	

	def change_visibility(label, value=None):
		if label == 'warmstart data':
			index = -1
		else:
			index = int(label)
		
		if value is None:
			value = not lines[index][0].get_visible()
		[l.set_visible(value) for l in lines[index]]
		plt.draw()

		
	def show_all(event):
		for label in range(num_iterations):
			if not lines[label][0].get_visible():
				change_visibility(label, False)
				check.set_active(label)

	def hide_all(event):
		for label in range(num_iterations):
			if lines[label][0].get_visible():
				change_visibility(label, True)
				check.set_active(label)
			
	check.on_clicked(change_visibility)
	none_button.on_clicked(hide_all)
	all_button.on_clicked(show_all)


	def update_hover_annotation(line, annotation=hover_annotation):
		xdata = line.get_xdata()
		ydata = line.get_ydata()
		idx = line.get_gid()        
		annotation.xy = (xdata[-1], ydata[-1])

		if not tool_tip_strings is None:
			annotation.set_text(tool_tip_strings[config_ids[idx]])
		else:
			annotation.set_text(str(config_ids[idx]))

		fig.canvas.draw_idle()

	def onpick1(event):

		line = event.artist

		if not line in all_lines:
			return
		
		gid = line.get_gid()

		if gid in permanent_annotations:
			# remove permanent annotation
			permanent_annotations[gid].set_visible(False)
			del permanent_annotations[gid]
			fig.canvas.draw_idle()

		else:
			# add a new annotation

			xdata = line.get_xdata()
			ydata = line.get_ydata()
			idx = line.get_gid()

			if not tool_tip_strings is None:
				text = tool_tip_strings[config_ids[idx]]
			else:
				text = str(config_ids[idx])
			
			permanent_annotations[gid] = ax.annotate(text, copy.deepcopy((xdata[-1],ydata[-1])),
							xytext=(20,20),textcoords="offset points",
							bbox=dict(boxstyle="round", fc="w", linewidth=3),
							arrowprops=dict(arrowstyle="->"))
			permanent_annotations[gid].draggable()
			fig.canvas.draw_idle()            

	def hover(event):
		vis = hover_annotation.get_visible()
		if event.inaxes == ax:
			active_lines = list(filter(lambda l: l.contains(event)[0], all_lines))
			if len(active_lines) > 0:
				hover_annotation.set_visible(True)
				update_hover_annotation(active_lines[0])
			elif vis:
				hover_annotation.set_visible(False)
				fig.canvas.draw_idle()



	fig.canvas.mpl_connect('pick_event', onpick1)
	fig.canvas.mpl_connect("motion_notify_event", hover)


	if show:
		plt.show()
	return(fig, ax, check, none_button, all_button)



"""
Example 2 - Local and Parallel (using threads)
==============================================

This example extends example 1 to multiple parallel workers.
For demonstrational purposes, the workers are started in separate threads.
This is probably not the most common use case but serves as a nice example.
The setup can be useful if the workers do their computations outside Python's Global Interpreter Lock, e.g. in some 3rd party extension in C/C++.
In that case, all workers can truely work in parallel.

"""
import logging
logging.basicConfig(level=logging.WARNING)

import argparse

import hpbandster.core.nameserver as hpns
import hpbandster.core.result as hpres

from hpbandster.optimizers import BOHB as BOHB
from hpbandster.examples.commons import MyWorker



parser = argparse.ArgumentParser(description='Example 1 - sequential and local execution.')
parser.add_argument('--min_budget',   type=float, help='Minimum budget used during the optimization.',    default=9)
parser.add_argument('--max_budget',   type=float, help='Maximum budget used during the optimization.',    default=243)
parser.add_argument('--n_iterations', type=int,   help='Number of iterations performed by the optimizer', default=4)
parser.add_argument('--n_workers', type=int,   help='Number of workers to run in parallel.', default=2)

args=parser.parse_args()


# Step 1: Start a nameserver (see example_1)
NS = hpns.NameServer(run_id='example2', host='127.0.0.1', port=None)
NS.start()

# Step 2: Start the workers
# Now we can instantiate the specified number of workers. To emphasize the effect,
# we introduce a sleep_interval of one second, which makes every function evaluation
# take a bit of time. Note the additional id argument that helps separating the
# individual workers. This is necessary because every worker uses its processes
# ID which is the same for all threads here.
workers=[]
for i in range(args.n_workers):
	w = MyWorker(sleep_interval = 0.5, nameserver='127.0.0.1',run_id='example2', id=i)
	w.run(background=True)
	workers.append(w)

# Step 3: Run an optimizer
# Now we can create an optimizer object and start the run.
# We add the min_n_workers argument to the run methods to make the optimizer wait
# for all workers to start. This is not mandatory, and workers can be added
# at any time, but if the timing of the run is essential, this can be used to
# synchronize all workers right at the start.
bohb = BOHB(  configspace = w.get_configspace(),
              run_id = 'example2',
              min_budget=args.min_budget, max_budget=args.max_budget
           )
res = bohb.run(n_iterations=args.n_iterations, min_n_workers=args.n_workers)

# Step 4: Shutdown
# After the optimizer run, we must shutdown the master and the nameserver.
bohb.shutdown(shutdown_workers=True)
NS.shutdown()

# Step 5: Analysis
# Each optimizer returns a hpbandster.core.result.Result object.
# It holds informations about the optimization run like the incumbent (=best) configuration.
# For further details about the Result object, see its documentation.
# Here we simply print out the best config and some statistics about the performed runs.
id2config = res.get_id2config_mapping()
incumbent = res.get_incumbent_id()

all_runs = res.get_all_runs()

print('Best found configuration:', id2config[incumbent]['config'])
print('A total of %i unique configurations where sampled.' % len(id2config.keys()))
print('A total of %i runs where executed.' % len(res.get_all_runs()))
print('Total budget corresponds to %.1f full function evaluations.'%(sum([r.budget for r in all_runs])/args.max_budget))
print('Total budget corresponds to %.1f full function evaluations.'%(sum([r.budget for r in all_runs])/args.max_budget))
print('The run took  %.1f seconds to complete.'%(all_runs[-1].time_stamps['finished'] - all_runs[0].time_stamps['started']))



import ConfigSpace
import numpy as np
import threading

from robo.models.lcnet import LCNet, get_lc_net

from hpbandster.core.base_config_generator import base_config_generator


def smoothing(lc):
    new_lc = []
    curr_best = np.inf
    for i in range(len(lc)):
        if lc[i] < curr_best:
            curr_best = lc[i]
        new_lc.append(curr_best)
    return new_lc


class LCNetWrapper(base_config_generator):
    def __init__(self,
                 configspace,
                 max_budget,
                 n_points=2000,
                 delta=1.0,
                 n_candidates=1024,
                 **kwargs):
        """
        Parameters:
        -----------

        directory: string
            where the results are logged
        logger: hpbandster.utils.result_logger_v??
            the logger to store the data, defaults to v1
        overwrite: bool
            whether or not existing data will be overwritten

        """

        super(LCNetWrapper, self).__init__(**kwargs)

        self.n_candidates = n_candidates
        self.model = LCNet(sampling_method="sghmc",
                           l_rate=np.sqrt(1e-4),
                           mdecay=.05,
                           n_nets=100,
                           burn_in=500,
                           n_iters=3000,
                           get_net=get_lc_net,
                           precondition=True)

        self.config_space = configspace
        self.max_budget = max_budget
        self.train = None
        self.train_targets = None
        self.n_points = n_points
        self.is_trained = False
        self.counter = 0
        self.delta = delta
        self.lock = threading.Lock()

    def get_config(self, budget):
        """
            function to sample a new configuration

            This function is called inside Hyperband to query a new configuration


            Parameters:
            -----------
            budget: float
                the budget for which this configuration is scheduled

            returns: config
                should return a valid configuration

        """
        self.lock.acquire()
        if not self.is_trained:
            c = self.config_space.sample_configuration().get_array()
        else:
            candidates = np.array([self.config_space.sample_configuration().get_array()
                                   for _ in range(self.n_candidates)])

            # We are only interested on the asymptotic value
            projected_candidates = np.concatenate((candidates, np.ones([self.n_candidates, 1])), axis=1)

            # Compute the upper confidence bound of the function at the asymptote
            m, v = self.model.predict(projected_candidates)

            ucb_values = m + self.delta * np.sqrt(v)
            print(ucb_values)
            # Sample a configuration based on the ucb values
            p = np.ones(self.n_candidates) * (ucb_values / np.sum(ucb_values))
            idx = np.random.choice(self.n_candidates, 1, False, p)

            c = candidates[idx][0]

        config = ConfigSpace.Configuration(self.config_space, vector=c)

        self.lock.release()
        return config.get_dictionary(), {}

    def new_result(self, job):
        """
            function to register finished runs

            Every time a run has finished, this function should be called
            to register it with the result logger. If overwritten, make
            sure to call this method from the base class to ensure proper
            logging.


            Parameters:
            -----------
            job_id: dict
                a dictionary containing all the info about the run
            job_result: dict
                contains all the results of the job, i.e. it's a dict with
                the keys 'loss' and 'info'

        """
        super().new_result(job)

        conf = ConfigSpace.Configuration(self.config_space, job.kwargs['config']).get_array()

        epochs = len(job.result["info"]["learning_curve"])
        budget = int(job.kwargs["budget"])

        t_idx = np.linspace(budget / epochs, budget, epochs) / self.max_budget
        x_new = np.repeat(conf[None, :], t_idx.shape[0], axis=0)

        x_new = np.concatenate((x_new, t_idx[:, None]), axis=1)

        # Smooth learning curve
        lc = smoothing(job.result["info"]["learning_curve"])

        # Flip learning curves since LC-Net wants increasing curves
        lc_new = [1 - y for y in lc]

        if self.train is None:
            self.train = x_new
            self.train_targets = lc_new
        else:
            self.train = np.append(self.train, x_new, axis=0)
            self.train_targets = np.append(self.train_targets, lc_new, axis=0)

        if self.counter >= self.n_points:

            self.lock.acquire()
            y_min = np.min(self.train_targets)
            y_max = np.max(self.train_targets)

            train_targets = (self.train_targets - y_min) / (y_max - y_min)

            self.model.train(self.train, train_targets)
            self.is_trained = True
            self.counter = 0
            self.lock.release()

        else:
            self.counter += epochs



"""
Worker for Example 5 - PyTorch
==============================

In this example implements a small CNN in PyTorch to train it on MNIST.
The configuration space shows the most common types of hyperparameters and
even contains conditional dependencies.
In this example implements a small CNN in Keras to train it on MNIST.
The configuration space shows the most common types of hyperparameters and
even contains conditional dependencies.

We'll optimise the following hyperparameters:

+-------------------------+----------------+-----------------+------------------------+
| Parameter Name          | Parameter type |  Range/Choices  | Comment                |
+=========================+================+=================+========================+
| Learning rate           |  float         | [1e-6, 1e-2]    | varied logarithmically |
+-------------------------+----------------+-----------------+------------------------+
| Optimizer               | categorical    | {Adam, SGD }    | discrete choice        |
+-------------------------+----------------+-----------------+------------------------+
| SGD momentum            |  float         | [0, 0.99]       | only active if         |
|                         |                |                 | optimizer == SGD       |
+-------------------------+----------------+-----------------+------------------------+
| Number of conv layers   | integer        | [1,3]           | can only take integer  |
|                         |                |                 | values 1, 2, or 3      |
+-------------------------+----------------+-----------------+------------------------+
| Number of filters in    | integer        | [4, 64]         | logarithmically varied |
| the first conf layer    |                |                 | integer values         |
+-------------------------+----------------+-----------------+------------------------+
| Number of filters in    | integer        | [4, 64]         | only active if number  |
| the second conf layer   |                |                 | of layers >= 2         |
+-------------------------+----------------+-----------------+------------------------+
| Number of filters in    | integer        | [4, 64]         | only active if number  |
| the third conf layer    |                |                 | of layers == 3         |
+-------------------------+----------------+-----------------+------------------------+
| Dropout rate            |  float         | [0, 0.9]        | standard continuous    |
|                         |                |                 | parameter              |
+-------------------------+----------------+-----------------+------------------------+
| Number of hidden units  | integer        | [8,256]         | logarithmically varied |
| in fully connected layer|                |                 | integer values         |
+-------------------------+----------------+-----------------+------------------------+

Please refer to the compute method below to see how those are defined using the
ConfigSpace package.
	  
The network does not achieve stellar performance when a random configuration is samples,
but a few iterations should yield an accuracy of >90%. To speed up training, only
8192 images are used for training, 1024 for validation.
The purpose is not to achieve state of the art on MNIST, but to show how to use
PyTorch inside HpBandSter, and to demonstrate a more complicated search space.
"""

try:
	import torch
	import torch.utils.data
	import torch.nn as nn
	import torch.nn.functional as F
except:
	raise ImportError("For this example you need to install pytorch.")

try:
	import torchvision
	import torchvision.transforms as transforms
except:
	raise ImportError("For this example you need to install pytorch-vision.")



import ConfigSpace as CS
import ConfigSpace.hyperparameters as CSH

from hpbandster.core.worker import Worker

import logging
logging.basicConfig(level=logging.DEBUG)



class PyTorchWorker(Worker):
	def __init__(self, N_train = 8192, N_valid = 1024, **kwargs):
		super().__init__(**kwargs)

		batch_size = 64

		# Load the MNIST Data here
		train_dataset = torchvision.datasets.MNIST(root='../../data', train=True, transform=transforms.ToTensor(), download=True)
		test_dataset = torchvision.datasets.MNIST(root='../../data', train=False, transform=transforms.ToTensor())
		
		train_sampler = torch.utils.data.sampler.SubsetRandomSampler(range(N_train))
		validation_sampler = torch.utils.data.sampler.SubsetRandomSampler(range(N_train, N_train+N_valid))

		
		self.train_loader = torch.utils.data.DataLoader(dataset=train_dataset, batch_size=batch_size, sampler=train_sampler)
		self.validation_loader = torch.utils.data.DataLoader(dataset=train_dataset, batch_size=1024, sampler=validation_sampler)

		self.test_loader = torch.utils.data.DataLoader(dataset=test_dataset, batch_size=1024, shuffle=False)


	def compute(self, config, budget, working_directory, *args, **kwargs):
		"""
		Simple example for a compute function using a feed forward network.
		It is trained on the MNIST dataset.
		The input parameter "config" (dictionary) contains the sampled configurations passed by the bohb optimizer
		"""

		# device = torch.device('cpu')
		model = MNISTConvNet(num_conv_layers=config['num_conv_layers'],
							num_filters_1=config['num_filters_1'],
							num_filters_2=config['num_filters_2'] if 'num_filters_2' in config else None,
							num_filters_3=config['num_filters_3'] if 'num_filters_3' in config else None,
							dropout_rate=config['dropout_rate'],
							num_fc_units=config['num_fc_units'],
							kernel_size=3
		)

		criterion = torch.nn.CrossEntropyLoss()
		if config['optimizer'] == 'Adam':
			optimizer = torch.optim.Adam(model.parameters(), lr=config['lr'])
		else:
			optimizer = torch.optim.SGD(model.parameters(), lr=config['lr'], momentum=config['sgd_momentum'])

		for epoch in range(int(budget)):
			loss = 0
			model.train()
			for i, (x, y) in enumerate(self.train_loader):
				optimizer.zero_grad()
				output = model(x)
				loss = F.nll_loss(output, y)
				loss.backward()
				optimizer.step()

		train_accuracy = self.evaluate_accuracy(model, self.train_loader)
		validation_accuracy = self.evaluate_accuracy(model, self.validation_loader)
		test_accuracy = self.evaluate_accuracy(model, self.test_loader)

		return ({
			'loss': 1-validation_accuracy, # remember: HpBandSter always minimizes!
			'info': {	'test accuracy': test_accuracy,
						'train accuracy': train_accuracy,
						'validation accuracy': validation_accuracy,
						'number of parameters': model.number_of_parameters(),
					}
						
		})

	def evaluate_accuracy(self, model, data_loader):
		model.eval()
		correct=0
		with torch.no_grad():
			for x, y in data_loader:
				output = model(x)
				#test_loss += F.nll_loss(output, target, reduction='sum').item() # sum up batch loss
				pred = output.max(1, keepdim=True)[1] # get the index of the max log-probability
				correct += pred.eq(y.view_as(pred)).sum().item()
		#import pdb; pdb.set_trace()	
		accuracy = correct/len(data_loader.sampler)
		return(accuracy)


	@staticmethod
	def get_configspace():
		"""
		It builds the configuration space with the needed hyperparameters.
		It is easily possible to implement different types of hyperparameters.
		Beside float-hyperparameters on a log scale, it is also able to handle categorical input parameter.
		:return: ConfigurationsSpace-Object
		"""
		cs = CS.ConfigurationSpace()

		lr = CSH.UniformFloatHyperparameter('lr', lower=1e-6, upper=1e-1, default_value='1e-2', log=True)

		# For demonstration purposes, we add different optimizers as categorical hyperparameters.
		# To show how to use conditional hyperparameters with ConfigSpace, we'll add the optimizers 'Adam' and 'SGD'.
		# SGD has a different parameter 'momentum'.
		optimizer = CSH.CategoricalHyperparameter('optimizer', ['Adam', 'SGD'])

		sgd_momentum = CSH.UniformFloatHyperparameter('sgd_momentum', lower=0.0, upper=0.99, default_value=0.9, log=False)

		cs.add_hyperparameters([lr, optimizer, sgd_momentum])

		# The hyperparameter sgd_momentum will be used,if the configuration
		# contains 'SGD' as optimizer.
		cond = CS.EqualsCondition(sgd_momentum, optimizer, 'SGD')
		cs.add_condition(cond)

		num_conv_layers =  CSH.UniformIntegerHyperparameter('num_conv_layers', lower=1, upper=3, default_value=2)
		
		num_filters_1 = CSH.UniformIntegerHyperparameter('num_filters_1', lower=4, upper=64, default_value=16, log=True)
		num_filters_2 = CSH.UniformIntegerHyperparameter('num_filters_2', lower=4, upper=64, default_value=16, log=True)
		num_filters_3 = CSH.UniformIntegerHyperparameter('num_filters_3', lower=4, upper=64, default_value=16, log=True)


		cs.add_hyperparameters([num_conv_layers, num_filters_1, num_filters_2, num_filters_3])
		
		# You can also use inequality conditions:
		cond = CS.GreaterThanCondition(num_filters_2, num_conv_layers, 1)
		cs.add_condition(cond)

		cond = CS.GreaterThanCondition(num_filters_3, num_conv_layers, 2)
		cs.add_condition(cond)


		dropout_rate = CSH.UniformFloatHyperparameter('dropout_rate', lower=0.0, upper=0.9, default_value=0.5, log=False)
		num_fc_units = CSH.UniformIntegerHyperparameter('num_fc_units', lower=8, upper=256, default_value=32, log=True)

		cs.add_hyperparameters([dropout_rate, num_fc_units])

		return cs




class MNISTConvNet(torch.nn.Module):
	def __init__(self, num_conv_layers, num_filters_1, num_filters_2, num_filters_3, dropout_rate, num_fc_units, kernel_size):
		super().__init__()
		
		self.conv1 = nn.Conv2d(1, num_filters_1, kernel_size=kernel_size)
		self.conv2 = None
		self.conv3 = None
		
		output_size = (28-kernel_size + 1)//2
		num_output_filters = num_filters_1
		
		if num_conv_layers > 1:
			self.conv2 = nn.Conv2d(num_filters_1, num_filters_2, kernel_size=kernel_size)
			num_output_filters = num_filters_2
			output_size = (output_size - kernel_size + 1)//2

		if num_conv_layers > 2:
			self.conv3 = nn.Conv2d(num_filters_2, num_filters_3, kernel_size=kernel_size)
			num_output_filters = num_filters_3
			output_size = (output_size - kernel_size + 1)//2
		
		self.dropout = nn.Dropout(p = dropout_rate)

		self.conv_output_size = num_output_filters*output_size*output_size

		self.fc1 = nn.Linear(self.conv_output_size, num_fc_units)
		self.fc2 = nn.Linear(num_fc_units, 10)
		


	def forward(self, x):
		
		# switched order of pooling and relu compared to the original example
		# to make it identical to the keras worker
		# seems to also give better accuracies
		x = F.max_pool2d(F.relu(self.conv1(x)), 2)
		
		if not self.conv2 is None:
			x = F.max_pool2d(F.relu(self.conv2(x)), 2)

		if not self.conv3 is None:
			x = F.max_pool2d(F.relu(self.conv3(x)), 2)

		x = self.dropout(x)
		
		x = x.view(-1, self.conv_output_size)
		x = F.relu(self.fc1(x))
		x = self.dropout(x)
		x = self.fc2(x)
		return F.log_softmax(x, dim=1)


	def number_of_parameters(self):
		return(sum(p.numel() for p in self.parameters() if p.requires_grad))



if __name__ == "__main__":
	worker = KerasWorker(run_id='0')
	cs = worker.get_configspace()
	
	config = cs.sample_configuration().get_dictionary()
	print(config)
	res = worker.compute(config=config, budget=2, working_directory='.')
	print(res)



"""
Example 6 -  Analysis of a Run
==============================

This example takes a run from example 5 and performs some analysis of it.
It shows how to get the best performing configuration, and its attributes.
More advanced analysis plots provide some insights into a run and the problem.

"""

import matplotlib.pyplot as plt
import hpbandster.core.result as hpres
import hpbandster.visualization as hpvis



# load the example run from the log files
result = hpres.logged_results_to_HBS_result('example_5_run/')

# get all executed runs
all_runs = result.get_all_runs()

# get the 'dict' that translates config ids to the actual configurations
id2conf = result.get_id2config_mapping()


# Here is how you get he incumbent (best configuration)
inc_id = result.get_incumbent_id()

# let's grab the run on the highest budget 
inc_runs = result.get_runs_by_id(inc_id)
inc_run = inc_runs[-1]


# We have access to all information: the config, the loss observed during
#optimization, and all the additional information
inc_loss = inc_run.loss
inc_config = id2conf[inc_id]['config']
inc_test_loss = inc_run.info['test accuracy']

print('Best found configuration:')
print(inc_config)
print('It achieved accuracies of %f (validation) and %f (test).'%(1-inc_loss, inc_test_loss))


# Let's plot the observed losses grouped by budget,
hpvis.losses_over_time(all_runs) 

# the number of concurent runs,
hpvis.concurrent_runs_over_time(all_runs) 

# and the number of finished runs.
hpvis.finished_runs_over_time(all_runs)

# This one visualizes the spearman rank correlation coefficients of the losses
# between different budgets.
hpvis.correlation_across_budgets(result)

# For model based optimizers, one might wonder how much the model actually helped.
# The next plot compares the performance of configs picked by the model vs. random ones
hpvis.performance_histogram_model_vs_random(all_runs, id2conf) 

plt.show()



"""
Example 7 - Interactive Exploration of the Results
==================================================

This example takes a run from example 5 and allows you to interactively look at
all runs performed and the associated loss.
The plot allows you to only include certain iterations (selected by the checkboxes on the left).
By hovering over the learning curves (all runs of a single configuration on all respective budets)
you see some information about the configuration and its performance.
Clicking on it will make the tool tip persistent.Clicking on the curve again, will remove it again.

This tool is not very mature, but maybe it can help you to explore the structure hidden in your
results.  Please refer to the documentation of the visualization submodule to see all
options.
"""

import matplotlib.pyplot as plt
import hpbandster.core.result as hpres
import hpbandster.visualization as hpvis



# load the example run from the log files
result = hpres.logged_results_to_HBS_result('example_5_run/')

# get all executed runs
all_runs = result.get_all_runs()

# get the 'dict' that translates config ids to the actual configurations
id2conf = result.get_id2config_mapping()

lcs = result.get_learning_curves()

hpvis.interactive_HBS_plot(lcs, tool_tip_strings=hpvis.default_tool_tips(result, lcs))




def realtime_learning_curves(runs):
	"""
	example how to extract a different kind of learning curve.
	
	The x values are now the time the runs finished, not the budget anymore.
	We no longer plot the validation loss on the y axis, but now the test accuracy.
	
	This is just to show how to get different information into the interactive plot.
	
	"""
	sr = sorted(runs, key=lambda r: r.budget)
	lc = list(filter(lambda t: not t[1] is None, [(r.time_stamps['finished'], r.info['test accuracy']) for r in sr]))
	return([lc,])
	
lcs = result.get_learning_curves(lc_extractor=realtime_learning_curves)

hpvis.interactive_HBS_plot(lcs, tool_tip_strings=hpvis.default_tool_tips(result, lcs))



"""
Worker for Examples 1-4
=======================

This class implements a very simple worker used in the firt examples.
"""

import numpy
import time

import ConfigSpace as CS
from hpbandster.core.worker import Worker


class MyWorker(Worker):

    def __init__(self, *args, sleep_interval=0, **kwargs):
        super().__init__(*args, **kwargs)

        self.sleep_interval = sleep_interval

    def compute(self, config, budget, **kwargs):
        """
        Simple example for a compute function
        The loss is just a the config + some noise (that decreases with the budget)

        For dramatization, the function can sleep for a given interval to emphasizes
        the speed ups achievable with parallel workers.

        Args:
            config: dictionary containing the sampled configurations by the optimizer
            budget: (float) amount of time/epochs/etc. the model can use to train

        Returns:
            dictionary with mandatory fields:
                'loss' (scalar)
                'info' (dict)
        """

        res = numpy.clip(config['x'] + numpy.random.randn()/budget, config['x']/2, 1.5*config['x'])
        time.sleep(self.sleep_interval)

        return({
                    'loss': float(res),  # this is the a mandatory field to run hyperband
                    'info': res  # can be used for any user-defined information - also mandatory
                })
    
    @staticmethod
    def get_configspace():
        config_space = CS.ConfigurationSpace()
        config_space.add_hyperparameter(CS.UniformFloatHyperparameter('x', lower=0, upper=1))
        return(config_space)




import os.path
import json
import threading

import Pyro4
import Pyro4.naming


from hpbandster.core.result import Result
from hpbandster.core.base_iteration import Datum



def nic_name_to_host(nic_name):
	""" translates the name of a network card into a valid host name"""
	from netifaces import ifaddresses, AF_INET
	host = ifaddresses(nic_name).setdefault(AF_INET, [{'addr': 'No IP addr'}] )[0]['addr']
	return(host)



def start_local_nameserver(host=None, port=0, nic_name=None):
	"""
		starts a Pyro4 nameserver in a daemon thread
		
		Parameters:
		-----------
			host: str
				the hostname to use for the nameserver
			port: int
				the port to be used. Default =0 means a random port
			nic_name: str
				name of the network interface to use
		
		Returns:
		--------
			tuple (str, int):
				the host name and the used port
	"""
	
	if host is None:
		if nic_name is None:
			host = 'localhost'
		else:
			host = nic_name_to_host(nic_name)

	uri, ns, _ = Pyro4.naming.startNS(host=host, port=port)
	host, port = ns.locationStr.split(':')
	
	
	thread = threading.Thread(target=ns.requestLoop, name='Pyro4 nameserver started by HpBandSter')
	thread.daemon=True
	
	thread.start()
	return(host, int(port))






import os
import time
import math
import copy
import logging

import numpy as np

import ConfigSpace as CS

from hpbandster.core.master import Master
from hpbandster.optimizers.iterations import SuccessiveHalving
from hpbandster.optimizers.config_generators.lcnet import LCNetWrapper as CG_LCNet
# from hpbandster.optimizers.config_generators import RandomSampling


class LCNet(Master):
    def __init__(self, configspace=None,
                 eta=3, min_budget=0.01, max_budget=1,
                 min_points_in_model=None, top_n_percent=15,
                 num_samples=64, random_fraction=1 / 3, bandwidth_factor=3,
                 min_bandwidth=1e-3, n_points=200, n_candidates=1024, delta=1,
                 **kwargs):
        """

        Parameters
        ----------
        configspace: ConfigSpace object
            valid representation of the search space
        eta : float
            In each iteration, a complete run of sequential halving is executed. In it,
            after evaluating each configuration on the same subset size, only a fraction of
            1/eta of them 'advances' to the next round.
            Must be greater or equal to 2.
        min_budget : float
            The smallest budget to consider. Needs to be positive!
        max_budget : float
            the largest budget to consider. Needs to be larger than min_budget!
            The budgets will be geometrically distributed $\sim \eta^k$ for
            $k\in [0, 1, ... , num_subsets - 1]$.
        min_points_in_model: int
            number of observations to start building a KDE. Default 'None' means
            dim+1, the bare minimum.
        top_n_percent: int
            percentage ( between 1 and 99, default 15) of the observations that are considered good.
        num_samples: int
            number of samples to optimize EI (default 64)
        random_fraction: float
            fraction of purely random configurations that are sampled from the
            prior without the model.
        bandwidth_factor: float
            to encourage diversity, the points proposed to optimize EI, are sampled
            from a 'widened' KDE where the bandwidth is multiplied by this factor (default: 3)
        min_bandwidth: float
            to keep diversity, even when all (good) samples have the same value for one of the parameters,
            a minimum bandwidth (Default: 1e-3) is used instead of zero.
        iteration_kwargs: dict
            kwargs to be added to the instantiation of each iteration
        """

        # TODO: Propper check for ConfigSpace object!
        if configspace is None:
            raise ValueError("You have to provide a valid CofigSpace object")

        cg = CG_LCNet(configspace=configspace,
                      max_budget=max_budget,
                      n_points=n_points,
                      n_candidates=n_candidates,
                      delta=delta)

        # cg = RandomSampling(configspace)

        super().__init__(config_generator=cg, **kwargs)

        # Hyperband related stuff
        self.eta = eta
        self.min_budget = min_budget
        self.max_budget = max_budget

        # precompute some HB stuff
        self.max_SH_iter = -int(np.log(min_budget / max_budget) / np.log(eta)) + 1
        self.budgets = max_budget * np.power(eta, -np.linspace(self.max_SH_iter - 1, 0, self.max_SH_iter))

        self.config.update({
            'eta': eta,
            'min_budget': min_budget,
            'max_budget': max_budget,
            'budgets': self.budgets,
            'max_SH_iter': self.max_SH_iter,
            'min_points_in_model': min_points_in_model,
            'top_n_percent': top_n_percent,
            'num_samples': num_samples,
            'random_fraction': random_fraction,
            'bandwidth_factor': bandwidth_factor,
            'min_bandwidth': min_bandwidth
        })

    def get_next_iteration(self, iteration, iteration_kwargs={}):
        """
            BO-HB uses (just like Hyperband) SuccessiveHalving for each iteration.
            See Li et al. (2016) for reference.

            Parameters:
            -----------
                iteration: int
                    the index of the iteration to be instantiated

            Returns:
            --------
                SuccessiveHalving: the SuccessiveHalving iteration with the
                    corresponding number of configurations
        """

        # number of 'SH rungs'
        s = self.max_SH_iter - 1 - (iteration % self.max_SH_iter)
        # number of configurations in that bracket
        n0 = int(np.floor((self.max_SH_iter) / (s + 1)) * self.eta ** s)
        ns = [max(int(n0 * (self.eta ** (-i))), 1) for i in range(s + 1)]

        return (SuccessiveHalving(HPB_iter=iteration, num_configs=ns, budgets=self.budgets[(-s - 1):],
                                  config_sampler=self.config_generator.get_config, **iteration_kwargs))



import numpy as np
import scipy.optimize as spo

import ConfigSpace as CS

from hpbandster.optimizers.kde.kernels import Gaussian, AitchisonAitken, WangRyzinOrdinal, WangRyzinInteger


class MultivariateKDE(object):
	def __init__(self, configspace, fully_dimensional=True, min_bandwidth=1e-4, fix_boundary=True):
		"""
		Parameters:
		-----------
			configspace: ConfigSpace.ConfigurationSpace object
				description of the configuration space
			fully_dimensional: bool
				if True, a true multivariate KDE is build, otherwise it's approximated by
				the product of one dimensional KDEs
				
			min_bandwidth: float
				a lower limit to the bandwidths which can insure 'uncertainty'
		
		"""
		self.configspace = configspace
		self.types, self.num_values = self._get_types()
		self.min_bandwidth=min_bandwidth
		self.fully_dimensional=fully_dimensional
		self.fix_boundary = fix_boundary
		
		
		# precompute bandwidth bounds
		self.bw_bounds = []
		
		max_bw_cont=0.5
		max_bw_cat = 0.999
		
		for t in self.types:
			if t == 'C':
				self.bw_bounds.append((min_bandwidth, max_bw_cont))
			else:
				self.bw_bounds.append((min_bandwidth, max_bw_cat))
		
		self.bw_clip = np.array([ bwb[1] for bwb in self.bw_bounds ])

		# initialize other vars
		self.bandwidths = np.array([float('NaN')]*len(self.types))
		self.kernels = []
		for t,n in zip(self.types, self.num_values):
			
			kwargs = {'num_values':n, 'fix_boundary':fix_boundary}
			
			if t == 'I':	self.kernels.append(WangRyzinInteger(**kwargs))
			if t == 'C':	self.kernels.append(Gaussian(**kwargs))
			if t == 'O':	self.kernels.append(WangRyzinOrdinal(**kwargs))
			if t == 'U':	self.kernels.append(AitchisonAitken(**kwargs))			
		self.data = None
		
	
	
	def fit(self, data, weights=None, bw_estimator='scott', efficient_bw_estimation=True, update_bandwidth=True):
		"""
			fits the KDE to the data by estimating the bandwidths and storing the data
			
			Parameters
			----------
				data: 2d-array, shape N x M
					N datapoints in an M dimensional space to which the KDE is fit
				weights: 1d array
					N weights, one for every data point.
					They will be normalized to sum up to one
				fix_boundary_effects: bool
					whether to reweigh points close to the bondary no fix the pdf
				bw_estimator: str
					allowed values are 'scott' and 'mlcv' for Scott's rule of thumb
					and the maximum likelihood via cross-validation
				efficient_bw_estimation: bool
					if true, start bandwidth optimization from the previous value, otherwise
					start from Scott's values
				update_bandwidths: bool
					whether to update the bandwidths at all
		"""
		
		if self.data is None:
			# overwrite some values in case this is the first fit of the KDE
			efficient_bw_estimation = False
			update_bandwidth=True

		self.data = np.asfortranarray(data)
		for i,k in enumerate(self.kernels):
				self.kernels[i].data = self.data[:,i]

		self.weights = self._normalize_weights(weights)

		if not update_bandwidth:
			return
		
		if not efficient_bw_estimation or bw_estimator == 'scott':
			# inspired by the the statsmodels code
			sigmas = np.std(self.data, ddof=1, axis=0)
			IQRs = np.subtract.reduce(np.percentile(self.data, [75,25], axis=0))
			self.bandwidths = 1.059 * np.minimum(sigmas, IQRs) * np.power(self.data.shape[0], -0.2)
			# crop bandwidths for categorical parameters
			self.bandwidths = np.clip(self.bandwidths , self.min_bandwidth, self.bw_clip)
			
		if bw_estimator == 'mlcv':
			# optimize bandwidths here
			def opt_me(bw):
				self.bandwidths=bw
				self._set_kernel_bandwidths()
				return(self.loo_negloglikelihood())
			
			res = spo.minimize(opt_me, self.bandwidths, bounds=self.bw_bounds, method='SLSQP')
			self.optimizer_result = res
			self.bandwidths[:] = res.x
		self._set_kernel_bandwidths()

	def _set_kernel_bandwidths(self):
		for i,b in enumerate(self.bandwidths):
			self.kernels[i].set_bandwidth(b)


	def set_bandwidths(self, bandwidths):
		self.bandwidths[:] = bandwidths
		self._set_kernel_bandwidths()

	def _normalize_weights(self, weights):
		
		weights = np.ones(self.data.shape[0]) if weights is None else weights
		weights /= weights.sum()
		
		return(weights)


	def _individual_pdfs(self, x_test):
	
		pdfs = np.zeros(shape=[x_test.shape[0], self.data.shape[0], self.data.shape[1]], dtype=np.float)

		for i, k in enumerate(self.kernels):
			pdfs[:,:,i] = k(x_test[:,i]).T
		
		return(pdfs)


	def loo_negloglikelihood(self):
		# get all pdf values of the training data (including 'self interaction')
		pdfs = self._individual_pdfs(self.data)
		
		# get indices to remove diagonal values for LOO part :)
		indices = np.diag_indices(pdfs.shape[0])

		# combine values based on fully_dimensional!
		if self.fully_dimensional:
			pdfs[indices] = 0 # remove self interaction		
			
			pdfs2 = np.sum(np.prod(pdfs, axis=-1), axis=-1)
			sm_return_value = -np.log(pdfs2).sum()
			pdfs = np.prod(pdfs, axis=-1)
			
			# take weighted average (accounts for LOO!)
			lhs = np.sum(pdfs*self.weights, axis=-1)/(1-self.weights)
		else:
			#import pdb; pdb.set_trace()
			pdfs[indices] = 0 # we sum first so 0 is the appropriate value
			pdfs *= self.weights[:,None,None]
			
			pdfs = pdfs.sum(axis=-2)/(1-self.weights[:,None])
			lhs = np.prod(pdfs, axis=-1)
			
		return(-np.sum(self.weights*np.log(lhs)))


	def pdf(self, x_test):
		"""
			Computes the probability density function at all x_test
		"""
		N,D = self.data.shape
		x_test = np.asfortranarray(x_test)
		x_test = x_test.reshape([-1, D])
		
		pdfs = self._individual_pdfs(x_test)
		#import pdb; pdb.set_trace()
		# combine values based on fully_dimensional!
		if self.fully_dimensional:
			# first the product of the individual pdfs for each point in the data across dimensions and then the average (factorized kernel)
			pdfs = np.sum(np.prod(pdfs, axis=-1)*self.weights[None, :], axis=-1)
		else:
			# first the average over the 1d pdfs and the the product over dimensions (TPE like factorization of the pdf)
			pdfs = np.prod(np.sum(pdfs*self.weights[None,:,None], axis=-2), axis=-1)
		return(pdfs)


	def sample(self, num_samples=1):
		
		
		samples = np.zeros([num_samples, len(self.types)], dtype=np.float)

		if self.fully_dimensional:
			sample_indices = np.random.choice(self.data.shape[0], size=num_samples)
			
		else:
			sample_indices=None

		for i,k in enumerate(self.kernels):
			samples[:,i] = k.sample(sample_indices, num_samples)


		return(samples)
		

	def _get_types(self):
		""" extracts the needed types from the configspace for faster retrival later
		
			type = 0 - numerical (continuous or integer) parameter
			type >=1 - categorical parameter
			
			TODO: figure out a way to properly handle ordinal parameters
		
		"""
		types = []
		num_values = []
		for hp in self.configspace.get_hyperparameters():
			#print(hp)
			if isinstance(hp, CS.CategoricalHyperparameter):
				types.append('U')
				num_values.append(len(hp.choices))
			elif isinstance(hp, CS.UniformIntegerHyperparameter):
				types.append('I')
				num_values.append((hp.upper - hp.lower + 1))
			elif isinstance(hp, CS.UniformFloatHyperparameter):
				types.append('C')
				num_values.append(np.inf)
			elif isinstance(hp, CS.OrdinalHyperparameter):
				types.append('O')
				num_values.append(len(hp.sequence))
			else:
				raise ValueError('Unsupported Parametertype %s'%type(hp))
		return(types, num_values)



import ConfigSpace
import numpy as np
import scipy.stats as sps
import statsmodels.api as sm

from hpbandster.config_generators.base import base_config_generator


class KernelDensityEstimator(base_config_generator):
	
	def __init__(self, configspace, top_n_percent=10, update_after_n_points=50,
				 min_points_in_model = None,
				 *kwargs):
		"""
			Fits for each given budget a kernel density estimator on the best N percent of the
			evaluated configurations on this budget.


			Parameters:
			-----------
			configspace: ConfigSpace
				Configuration space object
			top_n_percent: int
				Determines the percentile of configurations that will be used as training data
				for the kernel density estimator, e.g if set to 10 the 10% best configurations will be considered
				for training.
			update_after_n_points: int
				Specifies after how many new observed points the kernel density will be retrained.
			min_points_in_model: int
				minimum number of datapoints needed to fit a model

		"""
		super(KernelDensityEstimator, self).__init__(**kwargs)

		self.top_n_percent = top_n_percent
		self.update_after_n_points = update_after_n_points
		self.configspace = configspace
		
		self.min_points_in_model = min_points_in_model
		if min_points_in_model is None:
			self.min_points_in_model = len(self.configspace.get_hyperparameters())+1


		# TODO: so far we only consider continuous configuration spaces
		self.var_type = "c" * len(self.configspace.get_hyperparameters())
		self.configs = dict()
		self.losses = dict()
		self.kde_models = dict()
		
	def get_config(self, budget):
		"""
			Function to sample a new configuration

			This function is called inside Hyperband to query a new configuration


			Parameters:
			-----------
			budget: float
				the budget for which this configuration is scheduled

			returns: config
				should return a valid configuration

		"""
		# No observations available for this budget sample from the prior
		if len(self.kde_models.keys()) == 0:
			return self.configspace.sample_configuration().get_dictionary()
		# If we haven't seen anything with this budget, we sample from the kde trained on the highest budget
		if budget not in self.kde_models.keys():
			budget = sorted(self.kde_models.keys())[-1]
		# TODO: This only works in continuous space and with gaussian kernels
		kde = self.kde_models[budget]
		idx = np.random.randint(0, len(self.kde_models[budget].data))

		vector = [sps.truncnorm.rvs(-m/bw,(1-m)/bw, loc=m, scale=bw) for m,bw in zip(self.kde_models[budget].data[idx], kde.bw)]
		
		if np.any(np.array(vector)>1) or np.any(np.array(vector)<0):
			raise RuntimeError("truncated normal sampling problems!")
		
		sample = ConfigSpace.Configuration(self.configspace, vector=vector)
		return sample.get_dictionary(), {}

	def new_result(self, job):
		"""
			function to register finished runs

			Every time a run has finished, this function should be called
			to register it with the result logger. If overwritten, make
			sure to call this method from the base class to ensure proper
			logging.


			Parameters:
			-----------
			job_id: dict
				a dictionary containing all the info about the run
			job_result: dict
				contains all the results of the job, i.e. it's a dict with
				the keys 'loss' and 'info'

		"""

		super(KernelDensityEstimator, self).new_result(job)
		budget = job.kwargs["budget"]
		if budget not in self.configs.keys():
			self.configs[budget] = []
			self.losses[budget] = []

		# We want to get a numerical representation of the configuration in the original space
		conf = ConfigSpace.Configuration(self.configspace, job.kwargs['config'])
		self.configs[budget].append(conf.get_array())
		self.losses[budget].append(job.result['result']["loss"])


		# Check if we have enough data points to fit a KDE
		if len(self.configs[budget]) % self.update_after_n_points == 0:
			train_configs, train_losses = [], []
		
			train_configs.extend(self.configs[budget])
			train_losses.extend(self.losses[budget])

			n = int(self.top_n_percent * len(train_configs) / 100.)
			
			remaining_budgets = list(self.configs.keys())
			remaining_budgets.remove(budget)
			remaining_budgets.sort(reverse=True)


			for b in remaining_budgets:
				if  n >= self.min_points_in_model: break
				train_configs.extend(self.configs[b])
				train_losses.extend(self.losses[b])
				n = int(self.top_n_percent * len(train_configs) / 100.)

			if  len(train_losses) < self.min_points_in_model:
				return
			
			n = max(self.min_points_in_model, n)
			
			# Refit KDE for the current budget
			idx = np.argsort(train_losses)

			train_data = (np.array(train_configs)[idx])[:n]
			self.kde_models[budget] = sm.nonparametric.KDEMultivariate(data=train_data,
																 var_type=self.var_type,
																 bw='cv_ls')



"""
Example 1 - Local and Sequential
================================

"""
import logging
logging.basicConfig(level=logging.WARNING)

import argparse

import hpbandster.core.nameserver as hpns
import hpbandster.core.result as hpres

from hpbandster.optimizers import BOHB as BOHB
from hpbandster.examples.commons import MyWorker



parser = argparse.ArgumentParser(description='Example 1 - sequential and local execution.')
parser.add_argument('--min_budget',   type=float, help='Minimum budget used during the optimization.',    default=9)
parser.add_argument('--max_budget',   type=float, help='Maximum budget used during the optimization.',    default=243)
parser.add_argument('--n_iterations', type=int,   help='Number of iterations performed by the optimizer', default=4)
args=parser.parse_args()


# Step 1: Start a nameserver
# Every run needs a nameserver. It could be a 'static' server with a
# permanent address, but here it will be started for the local machine with the default port.
# The nameserver manages the concurrent running workers across all possible threads or clusternodes.
# Note the run_id argument. This uniquely identifies a run of any HpBandSter optimizer.
NS = hpns.NameServer(run_id='example1', host='127.0.0.1', port=None)
NS.start()

# Step 2: Start a worker
# Now we can instantiate a worker, providing the mandatory information
# Besides the sleep_interval, we need to define the nameserver information and
# the same run_id as above. After that, we can start the worker in the background,
# where it will wait for incoming configurations to evaluate.
w = MyWorker(sleep_interval = 0, nameserver='127.0.0.1',run_id='example1')
w.run(background=True)

# Step 3: Run an optimizer
# Now we can create an optimizer object and start the run.
# Here, we run BOHB, but that is not essential.
# The run method will return the `Result` that contains all runs performed.
bohb = BOHB(  configspace = w.get_configspace(),
              run_id = 'example1', nameserver='127.0.0.1',
              min_budget=args.min_budget, max_budget=args.max_budget
           )
res = bohb.run(n_iterations=args.n_iterations)

# Step 4: Shutdown
# After the optimizer run, we must shutdown the master and the nameserver.
bohb.shutdown(shutdown_workers=True)
NS.shutdown()

# Step 5: Analysis
# Each optimizer returns a hpbandster.core.result.Result object.
# It holds informations about the optimization run like the incumbent (=best) configuration.
# For further details about the Result object, see its documentation.
# Here we simply print out the best config and some statistics about the performed runs.
id2config = res.get_id2config_mapping()
incumbent = res.get_incumbent_id()

print('Best found configuration:', id2config[incumbent]['config'])
print('A total of %i unique configurations where sampled.' % len(id2config.keys()))
print('A total of %i runs where executed.' % len(res.get_all_runs()))
print('Total budget corresponds to %.1f full function evaluations.'%(sum([r.budget for r in res.get_all_runs()])/args.max_budget))



import os
import time
import math
import copy
import logging

import numpy as np


import ConfigSpace as CS

from hpbandster.core.master import Master
from hpbandster.optimizers.iterations import SuccessiveHalving
from hpbandster.optimizers.config_generators.random_sampling import RandomSampling as RS

class RandomSearch(Master):
	def __init__(self, configspace = None,
					eta = 3, min_budget=1, max_budget=1,
					**kwargs
					):
		"""
                Implements a random search across the search space for comparison.
                Candidates are sampled at random and run on the maximum budget.

		Parameters
		----------
		configspace: ConfigSpace object
			valid representation of the search space
		eta : float
			In each iteration, a complete run of sequential halving is executed. In it,
			after evaluating each configuration on the same subset size, only a fraction of
			1/eta of them 'advances' to the next round.
			Must be greater or equal to 2.
		budget : float
			budget for the evaluation
		"""

		# TODO: Propper check for ConfigSpace object!
		if configspace is None:
			raise ValueError("You have to provide a valid ConfigSpace object")



		cg = RS( configspace = configspace )

		super().__init__(config_generator=cg, **kwargs)

		# Hyperband related stuff
		self.eta = eta
		self.min_budget = max_budget
		self.max_budget = max_budget
		
		
		# precompute some HB stuff
		self.max_SH_iter = -int(np.log(min_budget/max_budget)/np.log(eta)) + 1
		self.budgets = max_budget * np.power(eta, -np.linspace(self.max_SH_iter-1, 0, self.max_SH_iter))

		# max total budget for one iteration
		self.budget_per_iteration = sum([b*self.eta**i for i, b in enumerate(self.budgets[::-1])])
		
		self.config.update({
						'eta'        : eta,
						'min_budget' : max_budget,
						'max_budget' : max_budget,
					})

	def get_next_iteration(self, iteration, iteration_kwargs={}):
		"""
		Returns a SH iteration with only evaluations on the biggest budget
		
		Parameters
		----------
			iteration: int
				the index of the iteration to be instantiated

		Returns
		-------
			SuccessiveHalving: the SuccessiveHalving iteration with the
				corresponding number of configurations
		"""
		
		
		budgets = [self.max_budget]
		ns = [self.budget_per_iteration//self.max_budget]
		
		return(SuccessiveHalving(HPB_iter=iteration, num_configs=ns, budgets=budgets, config_sampler=self.config_generator.get_config, **iteration_kwargs))



import os
import time
import math
import pdb
import copy
import logging

import numpy as np


import ConfigSpace as CS

from hpbandster.core.master import Master
from hpbandster.optimizers.iterations import SuccessiveHalving
from hpbandster.optimizers.config_generators import RandomSampling

class HyperBand(Master):
	def __init__(self, configspace = None,
					eta=3, min_budget=0.01, max_budget=1,
					**kwargs ):
		"""
                Hyperband implements hyperparameter optimization by sampling
                candidates at random and "trying" them first, running them for
                a specific budget. The approach is iterative, promising
                candidates are run for a longer time, increasing the fidelity
                for their performance. While this is a very efficient racing
                approach, random sampling makes no use of the knowledge gained
                about the candidates during optimization.

		Parameters
		----------
		configspace: ConfigSpace object
			valid representation of the search space
		eta : float
			In each iteration, a complete run of sequential halving is executed. In it,
			after evaluating each configuration on the same subset size, only a fraction of
			1/eta of them 'advances' to the next round.
			Must be greater or equal to 2.
		min_budget : float
			The smallest budget to consider. Needs to be positive!
		max_budget : float
			the largest budget to consider. Needs to be larger than min_budget!
			The budgets will be geometrically distributed $\sim \eta^k$ for
			$k\in [0, 1, ... , num_subsets - 1]$.
		"""


		# TODO: Propper check for ConfigSpace object!
		if configspace is None:
			raise ValueError("You have to provide a valid CofigSpace object")

		super().__init__(config_generator=RandomSampling(configspace), **kwargs)

		# Hyperband related stuff
		self.eta = eta
		self.min_budget = min_budget
		self.max_budget = max_budget

		# precompute some HB stuff
		self.max_SH_iter = -int(np.log(min_budget/max_budget)/np.log(eta)) + 1
		self.budgets = max_budget * np.power(eta, -np.linspace(self.max_SH_iter-1, 0, self.max_SH_iter))

		self.config.update({
						'eta'        : eta,
						'min_budget' : min_budget,
						'max_budget' : max_budget,
						'budgets'    : self.budgets,
						'max_SH_iter': self.max_SH_iter,
					})



	def get_next_iteration(self, iteration, iteration_kwargs={}):
		"""
		Hyperband uses SuccessiveHalving for each iteration.
		See Li et al. (2016) for reference.
		
		Parameters
		----------
			iteration: int
				the index of the iteration to be instantiated

		Returns
		-------
			SuccessiveHalving: the SuccessiveHalving iteration with the
				corresponding number of configurations
		"""
		
		# number of 'SH rungs'
		s = self.max_SH_iter - 1 - (iteration%self.max_SH_iter)
		# number of configurations in that bracket
		n0 = int(np.floor((self.max_SH_iter)/(s+1)) * self.eta**s)
		ns = [max(int(n0*(self.eta**(-i))), 1) for i in range(s+1)]

		return(SuccessiveHalving(HPB_iter=iteration, num_configs=ns, budgets=self.budgets[(-s-1):], config_sampler=self.config_generator.get_config, **iteration_kwargs))



import numpy as np
import scipy.optimize as spo

import ConfigSpace as CS

from scipy.special import erf

from pdb import set_trace


class BaseKernel(object):
	def __init__(self, data=None, bandwidth=None, fix_boundary=False, num_values=None):
		
		self.data = data
		self.bw = bandwidth
		self.fix_boundary=fix_boundary
		
		if num_values is None:
			num_values = len(np.unique(data))
		self.num_values = num_values
		
		if not data is None:
			self.weights = self._compute_weights()	

	def set_bandwidth(self, bandwidth):
		self.bw = bandwidth
		self.weights = self._compute_weights()

	def _compute_weights(self):
		return(1.)

	def __call__(self, x_test):
		raise NotImplementedError

	def sample(self, sample_indices=None, num_samples=1):
		raise NotImplementedError


class Gaussian(BaseKernel):

	def _compute_weights(self):
		if not self.fix_boundary:
			return(1.)

		weights = np.zeros(self.data.shape[0])
		for i,d in enumerate(self.data):
			weights[i] = 2./(erf((1-d)/(np.sqrt(2)*self.bw)) + erf(d/(np.sqrt(2)*self.bw)))
		
		return(weights[:,None])
		
	def __call__(self, x_test):
		distances = x_test[None,:] - self.data[:,None]
		pdfs = np.exp(-0.5* np.power(distances/self.bw, 2))/(2.5066282746310002 * self.bw)
		
		# reweigh to compensate for boundaries
		pdfs *= self.weights
		
		return(pdfs)

	def sample(self, sample_indices=None, num_samples=1):
		""" returns samples according to the KDE
		
			Parameters
			----------
				sample_inices: list of ints
					Indices into the training data used as centers for the samples
				
				num_samples: int
					if samples_indices is None, this specifies how many samples
					are drawn.
				
		"""
		if sample_indices is None:
			sample_indices = np.random.choice(self.data.shape[0], size=num_samples)

		samples = self.data[sample_indices]

		delta = np.random.normal(size=num_samples)*self.bw
		samples += delta
		oob_idx  = np.argwhere(np.logical_or(samples > 1, samples < 0)).flatten()

		while len(oob_idx) > 0:
			samples[oob_idx] -= delta[oob_idx]		# revert move
			delta[oob_idx] = np.random.normal(size=len(oob_idx))*self.bw
			samples[oob_idx] += delta[oob_idx]
			oob_idx = oob_idx[np.argwhere(np.logical_or(samples[oob_idx] > 1, samples[oob_idx] < 0))].flatten()

		return(samples)

class AitchisonAitken(BaseKernel):
	def __call__(self, x_test):
		distances = np.rint(x_test[None,:] - self.data[:,None])

		idx = np.abs(distances) == 0
		distances[idx] = 1 - self.bw
		distances[~idx] = self.bw/(self.num_values-1)
		
		return(distances)


	def sample(self, sample_indices=None, num_samples=1):
		""" returns samples according to the KDE
		
			Parameters
			----------
				sample_inices: list of ints
					Indices into the training data used as centers for the samples
				
				num_samples: int
					if samples_indices is None, this specifies how many samples
					are drawn.
				
		"""
		if sample_indices is None:
			sample_indices = np.random.choice(self.data.shape[0], size=num_samples)
		samples = self.data[sample_indices]

		samples = samples.squeeze()
		
		if self.num_values == 1:
			# handle cases where there is only one value!
			return(samples)
		
		probs = self.bw * np.ones(self.num_values)/(self.num_values-1)
		probs[0] = 1-self.bw
		
		delta = np.random.choice(self.num_values, size=num_samples, p = probs)
		samples = np.mod(samples + delta, self.num_values)	

		return(samples)

class WangRyzinOrdinal(BaseKernel):
		
	def _compute_weights(self):
		if not self.fix_boundary:
			return(1.)
		np.zeros(self.data.shape[0])
		self.weights=1.
		x_test = np.arange(self.num_values)
		pdfs  = self.__call__(x_test)
		weights = 1./pdfs.sum(axis=1)[:,None]
		return(weights)
		
	def __call__(self, x_test):
		distances = x_test[None,:] - self.data[:,None]

		idx = np.abs(distances) < .1 # distances smaller than that are considered zero
		
		pdfs = np.zeros_like(distances, dtype=np.float)
		pdfs[idx] = (1-self.bw)
		pdfs[~idx] = 0.5*(1-self.bw) * np.power(self.bw, np.abs(distances[~idx]))
		# reweigh to compensate for boundaries
		pdfs *= self.weights

		return(pdfs)

	def sample(self, sample_indices=None, num_samples=1):
		""" returns samples according to the KDE
		
			Parameters
			----------
				sample_inices: list of ints
					Indices into the training data used as centers for the samples
				
				num_samples: int
					if samples_indices is None, this specifies how many samples
					are drawn.
				
		"""
		if sample_indices is None:
			sample_indices = np.random.choice(self.data.shape[0], size=num_samples)
			samples = self.data[sample_indices]

		possible_steps = np.arange(-self.num_values+1,self.num_values)
		idx = (np.abs(possible_steps) < 1e-2)
		
		ps = 0.5*(1-self.bw) * np.power(self.bw, np.abs(possible_steps))
		ps[idx] = (1-self.bw)
		ps /= ps.sum()
		
		delta = np.zeros_like(samples)
		oob_idx = np.arange(samples.shape[0])

		while len(oob_idx) > 0:
			samples[oob_idx] -= delta[oob_idx]		# revert move
			delta[oob_idx] = np.random.choice(possible_steps, size=len(oob_idx), p=ps)
			samples[oob_idx] += delta[oob_idx]
			#import pdb; pdb.set_trace()
			oob_idx = oob_idx[np.argwhere(np.logical_or(samples[oob_idx] > self.num_values-0.9, samples[oob_idx] < -0.1)).flatten()]
		return(np.rint(samples))

class WangRyzinInteger(BaseKernel):
	def _compute_weights(self):
		if not self.fix_boundary:
			return(1.)
		weights = np.zeros(self.data.shape[0], dtype=np.float)
		x_test = np.linspace(1/(2*self.num_values), 1-(1/(2*self.num_values)), self.num_values, endpoint=True)
		self.weights = 1.
		pdfs  = self.__call__(x_test)
		weights = 1./pdfs.sum(axis=1)[:,None]
		return(weights)
		
	def __call__(self, x_test):
		distances = (x_test[None,:] - self.data[:,None])
		
		pdfs = np.zeros_like(distances, dtype=np.float)

		idx = np.abs(distances) < 1/(3*self.num_values) # distances smaller than that are considered zero
		pdfs[idx] = (1-self.bw)
		pdfs[~idx] = 0.5*(1-self.bw) * np.power(self.bw, np.abs(distances[~idx])*self.num_values)
		# reweigh to compensate for boundaries
		pdfs *= self.weights
				
		return(pdfs)

	def sample(self, sample_indices=None, num_samples=1):
		""" returns samples according to the KDE
		
			Parameters
			----------
				sample_inices: list of ints
					Indices into the training data used as centers for the samples
				
				num_samples: int
					if samples_indices is None, this specifies how many samples
					are drawn.
				
		"""
		if sample_indices is None:
			sample_indices = np.random.choice(self.data.shape[0], size=num_samples)
		samples = self.data[sample_indices]

		possible_steps = np.arange(-self.num_values+1,self.num_values) / self.num_values
		ps = 0.5*(1-self.bw) * np.power(self.bw, np.abs(possible_steps))
		ps[self.num_values-1] = (1-self.bw)
		ps /= ps.sum()
		
		delta = np.zeros_like(samples)
		oob_idx = np.arange(samples.shape[0])

		while len(oob_idx) > 0:
			samples[oob_idx] -= delta[oob_idx]		# revert move
			delta[oob_idx] = np.random.choice(possible_steps, size=len(oob_idx), p=ps)
			samples[oob_idx] += delta[oob_idx]
			oob_idx = oob_idx[np.argwhere(np.logical_or(samples[oob_idx] > 1-1/(3*self.num_values), samples[oob_idx] < 1/(3*self.num_values))).flatten()]

		
		return(samples)



"""
Example 3 - Local and Parallel (using processes)
================================================

Getting closer to a distributed setup, this examples shows how to connect a nameserver, an optimizer and several workers running in different processes.
This would also allow true parallelism if the workers do all the computation in Python, such that the thread based paralelization of example 2 would not work.

"""
import logging
logging.basicConfig(level=logging.INFO)

import argparse

import hpbandster.core.nameserver as hpns
import hpbandster.core.result as hpres

from hpbandster.optimizers import BOHB as BOHB
from hpbandster.examples.commons import MyWorker



parser = argparse.ArgumentParser(description='Example 3 - Local and Parallel Execution.')
parser.add_argument('--min_budget',   type=float, help='Minimum budget used during the optimization.',    default=9)
parser.add_argument('--max_budget',   type=float, help='Maximum budget used during the optimization.',    default=243)
parser.add_argument('--n_iterations', type=int,   help='Number of iterations performed by the optimizer', default=4)
parser.add_argument('--n_workers', type=int,   help='Number of workers to run in parallel.', default=2)
parser.add_argument('--worker', help='Flag to turn this into a worker process', action='store_true')

args=parser.parse_args()



if args.worker:
	w = MyWorker(sleep_interval = 0.5, nameserver='127.0.0.1',run_id='example3')
	w.run(background=False)
	exit(0)

# Start a nameserver (see example_1)
NS = hpns.NameServer(run_id='example3', host='127.0.0.1', port=None)
NS.start()


# Run an optimizer (see example_2)
bohb = BOHB(  configspace = MyWorker.get_configspace(),
			  run_id = 'example3',
			  min_budget=args.min_budget, max_budget=args.max_budget
		   )
res = bohb.run(n_iterations=args.n_iterations, min_n_workers=args.n_workers)

# Step 4: Shutdown
# After the optimizer run, we must shutdown the master and the nameserver.
bohb.shutdown(shutdown_workers=True)
NS.shutdown()

# Step 5: Analysis
# Each optimizer returns a hpbandster.core.result.Result object.
# It holds informations about the optimization run like the incumbent (=best) configuration.
# For further details about the Result object, see its documentation.
# Here we simply print out the best config and some statistics about the performed runs.
id2config = res.get_id2config_mapping()
incumbent = res.get_incumbent_id()

all_runs = res.get_all_runs()

print('Best found configuration:', id2config[incumbent]['config'])
print('A total of %i unique configurations where sampled.' % len(id2config.keys()))
print('A total of %i runs where executed.' % len(res.get_all_runs()))
print('Total budget corresponds to %.1f full function evaluations.'%(sum([r.budget for r in all_runs])/args.max_budget))
print('Total budget corresponds to %.1f full function evaluations.'%(sum([r.budget for r in all_runs])/args.max_budget))
print('The run took  %.1f seconds to complete.'%(all_runs[-1].time_stamps['finished'] - all_runs[0].time_stamps['started']))




import ConfigSpace as CS
from hpbandster.core.worker import Worker

class HPOlib2Worker(Worker):
	def __init__(self,benchmark, configspace=None, budget_name='budget', budget_preprocessor=None, measure_test_loss=False, config_as_array=True, **kwargs):
		
		super().__init__(**kwargs)
		self.benchmark = benchmark
		
		if configspace is None:
			self.configspace = benchmark.get_configuration_space()
		else:
			self.configspace = configspace
		
		self.budget_name=budget_name

		if budget_preprocessor is None:
			self.budget_preprocessor = lambda b: b
		else:
			self.budget_preprocessor = budget_preprocessor

		self.config_as_array = config_as_array

		self.measure_test_loss = measure_test_loss

	def compute(self, config, budget, **kwargs):

		if self.config_as_array:
			c = CS.Configuration(self.configspace, values=config)
		else:
			c = config
		
		kwargs = {self.budget_name: self.budget_preprocessor(budget)}
		res = self.benchmark.objective_function(c, **kwargs)
		if self.measure_test_loss:
			del kwargs[self.budget_name]
			res['test_loss'] = self.benchmark.objective_function_test(c, **kwargs)['function_value']
		return({
			'loss': res['function_value'],
			'info': res
		})



class LCModel():
    """
        base class for simple learning curve models
    """
    
    def fit(self, times, losses, configs=None):
        """
            function to train the model on the observed data

            Parameters:
            -----------

            times: list
                list of numpy arrays of the timesteps for each curve
            losses: list
                list of numpy arrays of the loss (the actual learning curve)
            configs: list or None
                list of the configurations for each sample. Each element
                has to be a numpy array. Set to None, if no configuration
                information is available.
        """
        raise NotImplementedError()


    def predict_unseen(self, times, config):
        """
            predict the loss of an unseen configuration

            Parameters:
            -----------

            times: numpy array
                times where to predict the loss
            config: numpy array
                the numerical representation of the config

            Returns:
            --------
            
            mean and variance prediction at input times for the given config
        """
        raise NotImplementedError()


    def extend_partial(self, times, obs_times, obs_losses, config=None):
        """
            extends a partially observed curve

            Parameters:
            -----------

            times: numpy array
                times where to predict the loss
            obs_times: numpy array
                times where the curve has already been observed
            obs_losses: numpy array
                corresponding observed losses
            config: numpy array
                numerical reperesentation of the config; None if no config
                information is available
                
            Returns:
            --------
            
            mean and variance prediction at input times
                
                
        """

    



import os
import time
import math
import copy
import logging

import numpy as np


import ConfigSpace as CS

from hpbandster.core.master import Master
from hpbandster.optimizers.iterations import SuccessiveHalving
from hpbandster.optimizers.config_generators.h2bo import H2BO as CG_H2BO

class H2BO(Master):
	def __init__(self,
					configspace = None,
					eta=3, min_budget=0.01, max_budget=1,
					min_points_in_model = None,	top_n_percent=15,
					num_samples = 32, random_fraction=1/3, bandwidth_factor=1,
					min_bandwidth=1e-3,bw_estimator='scott',fully_dimensional=True,
					**kwargs
					):
		"""

		Parameters
		----------
		configspace: ConfigSpace object
			valid representation of the search space
		eta : float
			In each iteration, a complete run of sequential halving is executed. In it,
			after evaluating each configuration on the same subset size, only a fraction of
			1/eta of them 'advances' to the next round.
			Must be greater or equal to 2.
		min_budget : float
			The smallest budget to consider. Needs to be positive!
		max_budget : float
			the largest budget to consider. Needs to be larger than min_budget!
			The budgets will be geometrically distributed $\sim \eta^k$ for
			$k\in [0, 1, ... , num_subsets - 1]$.
		min_points_in_model: int
			number of observations to start building a KDE. Default 'None' means
			dim+1, the bare minimum.
		top_n_percent: int
			percentage ( between 1 and 99, default 15) of the observations that are considered good.
		num_samples: int
			number of samples to optimize EI (default 64)
		random_fraction: float
			fraction of purely random configurations that are sampled from the
			prior without the model.
		bw_estimator: str
			controls the way the bandwidths are estimator. For 'scott' a quick rule of thumb based
			on the empirical variance is used, for 'mlvc' the likelihood based on 
			leave on out cross validation is maximized.
		min_bandwidth: float
			to keep diversity, even when all (good) samples have the same value for one of the parameters,
			a minimum bandwidth (Default: 1e-3) is used instead of zero. 
		iteration_kwargs: dict
			kwargs to be added to the instantiation of each iteration
		fully_dimensional: bool
			if true, the KDE is uses factored kernel across all dimensions, otherwise the PDF is a product of 1d PDFs
		"""



		# TODO: Propper check for ConfigSpace object!
		if configspace is None:
			raise ValueError("You have to provide a valid CofigSpace object")



		cg = CG_H2BO( configspace = configspace,
					min_points_in_model = min_points_in_model,
					top_n_percent=top_n_percent,
					num_samples = num_samples,
					random_fraction=random_fraction,
					bw_estimator=bw_estimator,
					min_bandwidth = min_bandwidth,
					fully_dimensional=fully_dimensional
					)

		super().__init__(config_generator=cg, **kwargs)

		# Hyperband related stuff
		self.eta = eta
		self.min_budget = min_budget
		self.max_budget = max_budget
		
		
		# precompute some HB stuff
		self.max_SH_iter = -int(np.log(min_budget/max_budget)/np.log(eta)) + 1
		self.budgets = max_budget * np.power(eta, -np.linspace(self.max_SH_iter-1, 0, self.max_SH_iter))



		# max total budget for one iteration
		self.budget_per_iteration = sum([b*self.eta**i for i, b in enumerate(self.budgets[::-1])])
		
		self.config.update({
						'eta'        : eta,
						'min_budget' : min_budget,
						'max_budget' : max_budget,
						'budgets'    : self.budgets,
						'max_SH_iter': self.max_SH_iter,
						'min_points_in_model' : min_points_in_model,
						'top_n_percent' : top_n_percent,
						'num_samples' : num_samples,
						'random_fraction' : random_fraction,
						'min_bandwidth': min_bandwidth,
						'bw_estimator': bw_estimator,
						'fully_dimensional': fully_dimensional,
					})

	def get_next_iteration(self, iteration, iteration_kwargs={}):
		"""
			BO-HB uses (just like Hyperband) SuccessiveHalving for each iteration.
			See Li et al. (2016) for reference.
			
			Parameters:
			-----------
				iteration: int
					the index of the iteration to be instantiated

			Returns:
			--------
				SuccessiveHalving: the SuccessiveHalving iteration with the
					corresponding number of configurations
		"""
		
		min_budget = max( self.min_budget, self.config_generator.largest_budget_with_model())
		max_budget = self.max_budget
		eta = self.eta

		# precompute some HB stuff
		max_SH_iter = -int(np.log(min_budget/max_budget)/np.log(eta)) + 1
		budgets = max_budget * np.power(eta, -np.linspace(max_SH_iter-1, 0, max_SH_iter))

		
		# number of 'SH rungs'
		s = max_SH_iter - 1
		# number of configurations in that bracket
		n0 = int(np.floor((self.max_SH_iter)/(s+1)) * eta**s)
		ns = np.array([max(int(n0*(eta**(-i))), 1) for i in range(s+1)])
		
		while (ns * budgets[-s-1:]).sum() <= self.budget_per_iteration:
		    n0 += 1
		    ns = np.array([max(int(n0*(eta**(-i))), 1) for i in range(s+1)])
		
		n0 -= 1
		ns = np.array([max(int(n0*(eta**(-i))), 1) for i in range(s+1)])
		
		assert (ns * budgets[-s-1:]).sum() <= self.budget_per_iteration, 'Sampled iteration exceeds the budget per iteration!'
		
		return(SuccessiveHalving(HPB_iter=iteration, num_configs=ns, budgets=budgets, config_sampler=self.config_generator.get_config, **iteration_kwargs))



from hpbandster.optimizers.randomsearch import RandomSearch
from hpbandster.optimizers.hyperband import HyperBand
from hpbandster.optimizers.bohb import BOHB
from hpbandster.optimizers.h2bo import H2BO



import logging
from copy import deepcopy
import traceback


import ConfigSpace
import ConfigSpace.hyperparameters
import ConfigSpace.util
import numpy as np
import scipy.stats as sps
import scipy.optimize as spo
import statsmodels.api as sm

from hpbandster.core.base_config_generator import base_config_generator


class BOHB(base_config_generator):
	def __init__(self, configspace, min_points_in_model = None,
				 top_n_percent=15, num_samples = 64, random_fraction=1/3,
				 bandwidth_factor=3, min_bandwidth=1e-3,
				**kwargs):
		"""
			Fits for each given budget a kernel density estimator on the best N percent of the
			evaluated configurations on this budget.


			Parameters:
			-----------
			configspace: ConfigSpace
				Configuration space object
			top_n_percent: int
				Determines the percentile of configurations that will be used as training data
				for the kernel density estimator, e.g if set to 10 the 10% best configurations will be considered
				for training.
			min_points_in_model: int
				minimum number of datapoints needed to fit a model
			num_samples: int
				number of samples drawn to optimize EI via sampling
			random_fraction: float
				fraction of random configurations returned
			bandwidth_factor: float
				widens the bandwidth for contiuous parameters for proposed points to optimize EI
			min_bandwidth: float
				to keep diversity, even when all (good) samples have the same value for one of the parameters,
				a minimum bandwidth (Default: 1e-3) is used instead of zero. 

		"""
		super().__init__(**kwargs)
		self.top_n_percent=top_n_percent
		self.configspace = configspace
		self.bw_factor = bandwidth_factor
		self.min_bandwidth = min_bandwidth

		self.min_points_in_model = min_points_in_model
		if min_points_in_model is None:
			self.min_points_in_model = len(self.configspace.get_hyperparameters())+1
		
		if self.min_points_in_model < len(self.configspace.get_hyperparameters())+1:
			self.logger.warning('Invalid min_points_in_model value. Setting it to %i'%(len(self.configspace.get_hyperparameters())+1))
			self.min_points_in_model =len(self.configspace.get_hyperparameters())+1
		
		self.num_samples = num_samples
		self.random_fraction = random_fraction

		hps = self.configspace.get_hyperparameters()

		self.kde_vartypes = ""
		self.vartypes = []


		for h in hps:
			if hasattr(h, 'sequence'):
				raise RuntimeError('This version on BOHB does not support ordinal hyperparameters. Please encode %s as an integer parameter!'%(h.name))
			
			if hasattr(h, 'choices'):
				self.kde_vartypes += 'u'
				self.vartypes +=[ len(h.choices)]
			else:
				self.kde_vartypes += 'c'
				self.vartypes +=[0]
		
		self.vartypes = np.array(self.vartypes, dtype=int)

		# store precomputed probs for the categorical parameters
		self.cat_probs = []
		

		self.configs = dict()
		self.losses = dict()
		self.good_config_rankings = dict()
		self.kde_models = dict()


	def largest_budget_with_model(self):
		if len(self.kde_models) == 0:
			return(-float('inf'))
		return(max(self.kde_models.keys()))

	def get_config(self, budget):
		"""
			Function to sample a new configuration

			This function is called inside Hyperband to query a new configuration


			Parameters:
			-----------
			budget: float
				the budget for which this configuration is scheduled

			returns: config
				should return a valid configuration

		"""
		
		self.logger.debug('start sampling a new configuration.')
		

		sample = None
		info_dict = {}
		
		# If no model is available, sample from prior
		# also mix in a fraction of random configs
		if len(self.kde_models.keys()) == 0 or np.random.rand() < self.random_fraction:
			sample =  self.configspace.sample_configuration()
			info_dict['model_based_pick'] = False

		best = np.inf
		best_vector = None

		if sample is None:
			try:
				
				#sample from largest budget
				budget = max(self.kde_models.keys())

				l = self.kde_models[budget]['good'].pdf
				g = self.kde_models[budget]['bad' ].pdf
			
				minimize_me = lambda x: max(1e-32, g(x))/max(l(x),1e-32)
				
				kde_good = self.kde_models[budget]['good']
				kde_bad = self.kde_models[budget]['bad']

				for i in range(self.num_samples):
					idx = np.random.randint(0, len(kde_good.data))
					datum = kde_good.data[idx]
					vector = []
					
					for m,bw,t in zip(datum, kde_good.bw, self.vartypes):
						
						bw = max(bw, self.min_bandwidth)
						if t == 0:
							bw = self.bw_factor*bw
							try:
								vector.append(sps.truncnorm.rvs(-m/bw,(1-m)/bw, loc=m, scale=bw))
							except:
								self.logger.warning("Truncated Normal failed for:\ndatum=%s\nbandwidth=%s\nfor entry with value %s"%(datum, kde_good.bw, m))
								self.logger.warning("data in the KDE:\n%s"%kde_good.data)
						else:
							
							if np.random.rand() < (1-bw):
								vector.append(int(m))
							else:
								vector.append(np.random.randint(t))
					val = minimize_me(vector)

					if not np.isfinite(val):
						self.logger.warning('sampled vector: %s has EI value %s'%(vector, val))
						self.logger.warning("data in the KDEs:\n%s\n%s"%(kde_good.data, kde_bad.data))
						self.logger.warning("bandwidth of the KDEs:\n%s\n%s"%(kde_good.bw, kde_bad.bw))
						self.logger.warning("l(x) = %s"%(l(vector)))
						self.logger.warning("g(x) = %s"%(g(vector)))

						# right now, this happens because a KDE does not contain all values for a categorical parameter
						# this cannot be fixed with the statsmodels KDE, so for now, we are just going to evaluate this one
						# if the good_kde has a finite value, i.e. there is no config with that value in the bad kde, so it shouldn't be terrible.
						if np.isfinite(l(vector)):
							best_vector = vector
							break

					if val < best:
						best = val
						best_vector = vector

				if best_vector is None:
					self.logger.debug("Sampling based optimization with %i samples failed -> using random configuration"%self.num_samples)
					sample = self.configspace.sample_configuration().get_dictionary()
					info_dict['model_based_pick']  = False
				else:
					self.logger.debug('best_vector: {}, {}, {}, {}'.format(best_vector, best, l(best_vector), g(best_vector)))
					for i, hp_value in enumerate(best_vector):
						if isinstance(
							self.configspace.get_hyperparameter(
								self.configspace.get_hyperparameter_by_idx(i)
							),
							ConfigSpace.hyperparameters.CategoricalHyperparameter
						):
							best_vector[i] = int(np.rint(best_vector[i]))
					sample = ConfigSpace.Configuration(self.configspace, vector=best_vector).get_dictionary()
					
					try:
						sample = ConfigSpace.util.deactivate_inactive_hyperparameters(
									configuration_space=self.configspace,
									configuration=sample
									)
						info_dict['model_based_pick'] = True

					except Exception as e:
						self.logger.warning(("="*50 + "\n")*3 +\
								"Error converting configuration:\n%s"%sample+\
								"\n here is a traceback:" +\
								traceback.format_exc())
						raise(e)

			except:
				self.logger.warning("Sampling based optimization with %i samples failed\n %s \nUsing random configuration"%(self.num_samples, traceback.format_exc()))
				sample = self.configspace.sample_configuration()
				info_dict['model_based_pick']  = False


		try:
			sample = ConfigSpace.util.deactivate_inactive_hyperparameters(
				configuration_space=self.configspace,
				configuration=sample.get_dictionary()
			).get_dictionary()
		except Exception as e:
			self.logger.warning("Error (%s) converting configuration: %s -> "
								"using random configuration!",
								e,
								sample)
			sample = self.configspace.sample_configuration().get_dictionary()
		self.logger.debug('done sampling a new configuration.')
		return sample, info_dict



	def impute_conditional_data(self, array):

		return_array = np.empty_like(array)

		for i in range(array.shape[0]):
			datum = np.copy(array[i])
			nan_indices = np.argwhere(np.isnan(datum)).flatten()

			while (np.any(nan_indices)):
				nan_idx = nan_indices[0]
				valid_indices = np.argwhere(np.isfinite(array[:,nan_idx])).flatten()

				if len(valid_indices) > 0:
					# pick one of them at random and overwrite all NaN values
					row_idx = np.random.choice(valid_indices)
					datum[nan_indices] = array[row_idx, nan_indices]

				else:
					# no good point in the data has this value activated, so fill it with a valid but random value
					t = self.vartypes[nan_idx]
					if t == 0:
						datum[nan_idx] = np.random.rand()
					else:
						datum[nan_idx] = np.random.randint(t)

				nan_indices = np.argwhere(np.isnan(datum)).flatten()
			return_array[i,:] = datum
		return(return_array)

	def new_result(self, job, update_model=True):
		"""
			function to register finished runs

			Every time a run has finished, this function should be called
			to register it with the result logger. If overwritten, make
			sure to call this method from the base class to ensure proper
			logging.


			Parameters:
			-----------
			job: hpbandster.distributed.dispatcher.Job object
				contains all the info about the run
		"""

		super().new_result(job)

		if job.result is None:
			# One could skip crashed results, but we decided to
			# assign a +inf loss and count them as bad configurations
			loss = np.inf
		else:
			# same for non numeric losses.
			# Note that this means losses of minus infinity will count as bad!
			loss = job.result["loss"] if np.isfinite(job.result["loss"] else np.inf

		budget = job.kwargs["budget"]

		if budget not in self.configs.keys():
			self.configs[budget] = []
			self.losses[budget] = []

		# skip model building if we already have a bigger model
		if max(list(self.kde_models.keys()) + [-np.inf]) > budget:
			return

		# We want to get a numerical representation of the configuration in the original space

		conf = ConfigSpace.Configuration(self.configspace, job.kwargs["config"])
		self.configs[budget].append(conf.get_array())
		self.losses[budget].append(loss)

		
		# skip model building:
		#		a) if not enough points are available
		if len(self.configs[budget]) <= self.min_points_in_model-1:
			self.logger.debug("Only %i run(s) for budget %f available, need more than %s -> can't build model!"%(len(self.configs[budget]), budget, self.min_points_in_model+1))
			return

		#		b) during warnm starting when we feed previous results in and only update once
		if not update_model:
			return

		train_configs = np.array(self.configs[budget])
		train_losses =  np.array(self.losses[budget])

		n_good= max(self.min_points_in_model, (self.top_n_percent * train_configs.shape[0])//100 )
		#n_bad = min(max(self.min_points_in_model, ((100-self.top_n_percent)*train_configs.shape[0])//100), 10)
		n_bad = max(self.min_points_in_model, ((100-self.top_n_percent)*train_configs.shape[0])//100)

		# Refit KDE for the current budget
		idx = np.argsort(train_losses)

		train_data_good = self.impute_conditional_data(train_configs[idx[:n_good]])
		train_data_bad  = self.impute_conditional_data(train_configs[idx[n_good:n_good+n_bad]])

		if train_data_good.shape[0] <= train_data_good.shape[1]:
			return
		if train_data_bad.shape[0] <= train_data_bad.shape[1]:
			return
		
		#more expensive crossvalidation method
		#bw_estimation = 'cv_ls'

		# quick rule of thumb
		bw_estimation = 'normal_reference'

		bad_kde = sm.nonparametric.KDEMultivariate(data=train_data_bad,  var_type=self.kde_vartypes, bw=bw_estimation)
		good_kde = sm.nonparametric.KDEMultivariate(data=train_data_good, var_type=self.kde_vartypes, bw=bw_estimation)

		bad_kde.bw = np.clip(bad_kde.bw, self.min_bandwidth,None)
		good_kde.bw = np.clip(good_kde.bw, self.min_bandwidth,None)

		self.kde_models[budget] = {
				'good': good_kde,
				'bad' : bad_kde
		}

		# update probs for the categorical parameters for later sampling
		self.logger.debug('done building a new model for budget %f based on %i/%i split\nBest loss for this budget:%f\n\n\n\n\n'%(budget, n_good, n_bad, np.min(train_losses)))




import sys


import logging
import numpy as np
import pdb

from hpbandster.core.dispatcher import Job


class Datum(object):
	def __init__(self, config, config_info, results=None, time_stamps=None, exceptions=None, status='QUEUED', budget=0):
		self.config		= config
		self.config_info= config_info
		self.results	= results		if not results is None else {}
		self.time_stamps= time_stamps	if not time_stamps is None else {}
		self.exceptions	= exceptions	if not exceptions is None else {}
		self.status		= status
		self.budget		= budget

	def __repr__(self):
		return(\
			"\nconfig:{}\n".format(self.config) + \
			"config_info:\n{}\n"%self.config_info + \
			"losses:\n"
			'\t'.join(["{}: {}\t".format(k, v['loss']) for k,v in self.results.items()]) + \
			"time stamps: {}".format(self.time_stamps)
		)



class BaseIteration(object):
	"""
	Base class for various iteration possibilities. This decides what configuration should be run on what budget
	next. Typical choices are e.g. successive halving. Results from runs are processed and (depending on the
	implementations) determine the further development.
	"""
	def __init__(self, HPB_iter, num_configs, budgets, config_sampler, logger=None, result_logger=None):
		"""
		Parameters
		----------

		HPB_iter: int
			The current HPBandSter iteration index.
		num_configs: list of ints
			the number of configurations in each stage of SH
		budgets: list of floats
			the budget associated with each stage
		config_sample: callable
			a function that returns a valid configuration. Its only
			argument should be the budget that this config is first
			scheduled for. This might be used to pick configurations
			that perform best after this particular budget is exhausted
			to build a better autoML system.
		logger: a logger
		result_logger: hpbandster.api.results.util.json_result_logger object
			a result logger that writes live results to disk
		"""

		self.data = {}					# this holds all the configs and results of this iteration
		self.is_finished = False
		self.HPB_iter = HPB_iter
		self.stage = 0					# internal iteration, but different name for clarity
		self.budgets = budgets
		self.num_configs = num_configs
		self.actual_num_configs = [0]*len(num_configs)
		self.config_sampler = config_sampler
		self.num_running = 0
		self.logger=logger if not logger is None else logging.getLogger('hpbandster')
		self.result_logger = result_logger

	def add_configuration(self, config = None, config_info={}):
		"""
		function to add a new configuration to the current iteration
		
		Parameters
		----------
		
		config : valid configuration
			The configuration to add. If None, a configuration is sampled from the config_sampler
		config_info: dict
			Some information about the configuration that will be stored in the results
		"""
		
		if config is None:
			config, config_info = self.config_sampler(self.budgets[self.stage])
		
		if self.is_finished:
			raise RuntimeError("This HPBandSter iteration is finished, you can't add more configurations!")

		if self.actual_num_configs[self.stage] == self.num_configs[self.stage]:
			raise RuntimeError("Can't add another configuration to stage %i in HPBandSter iteration %i."%(self.stage, self.HPB_iter))

		config_id = (self.HPB_iter, self.stage, self.actual_num_configs[self.stage])

		self.data[config_id] = Datum(config=config, config_info=config_info, budget = self.budgets[self.stage])

		self.actual_num_configs[self.stage] += 1
		
		if not self.result_logger is None:
		    self.result_logger.new_config(config_id, config, config_info)
		
		return(config_id)

	def register_result(self, job, skip_sanity_checks=False):
		"""
		function to register the result of a job

		This function is called from HB_master, don't call this from
		your script.
		"""

		if self.is_finished:
			raise RuntimeError("This HB iteration is finished, you can't register more results!")

		config_id = job.id
		config = job.kwargs['config']
		budget = job.kwargs['budget']
		timestamps = job.timestamps
		result = job.result
		exception = job.exception
		
		d = self.data[config_id]
		
		if not skip_sanity_checks:
			assert d.config == config, 'Configurations differ!'
			assert d.status == 'RUNNING', "Configuration wasn't scheduled for a run."
			assert d.budget == budget, 'Budgets differ (%f != %f)!'%(self.data[config_id]['budget'], budget)

		d.time_stamps[budget] = timestamps
		d.results[budget] = result

		if (not job.result is None) and np.isfinite(result['loss']):
			d.status = 'REVIEW'
		else:
			d.status = 'CRASHED'

		d.exceptions[budget] = exception
		self.num_running -= 1
		
	def get_next_run(self):
		"""
		function to return the next configuration and budget to run.

		This function is called from HB_master, don't call this from
		your script.

		It returns None if this run of SH is finished or there are
		pending jobs that need to finish to progress to the next stage.

		If there are empty slots to be filled in the current SH stage
		(which never happens in the original SH version), a new
		configuration will be sampled and scheduled to run next.
		"""

		if self.is_finished:
			return(None)
		
		for k,v in self.data.items():
			if v.status == 'QUEUED':
				assert v.budget == self.budgets[self.stage], 'Configuration budget does not align with current stage!'
				v.status = 'RUNNING'
				self.num_running += 1
				return(k, v.config, v.budget)

		# check if there are still slots to fill in the current stage and return that
		if (self.actual_num_configs[self.stage] < self.num_configs[self.stage]):
			self.add_configuration()
			return(self.get_next_run())

		if self.num_running == 0:
			# at this point a stage is completed
			self.process_results()
			return(self.get_next_run())

		return(None)

	def _advance_to_next_stage(self, config_ids, losses):
		"""
		Function that implements the strategy to advance configs within this iteration
		
		Overload this to implement different strategies, like
		SuccessiveHalving, SuccessiveResampling.
		
		Parameters
		----------
			config_ids: list
				all config ids to be considered
			losses: list
				losses of the run on the current budget
				
		Returns
		-------
			list of bool
				A boolean for each entry in config_ids indicating whether to advance it or not
		
		
		"""
		raise NotImplementedError('_advance_to_next_stage not implemented for %s'%type(self).__name__)

	def process_results(self):
		"""
		function that is called when a stage is completed and
		needs to be analyzed befor further computations.

		The code here implements the original SH algorithms by
		advancing the k-best (lowest loss) configurations at the current
		budget. k is defined by the num_configs list (see __init__)
		and the current stage value.

		For more advanced methods like resampling after each stage,
		overload this function only.
		"""
		self.stage += 1
		
		# collect all config_ids that need to be compared
		config_ids = list(filter(lambda cid: self.data[cid].status == 'REVIEW', self.data.keys()))

		if (self.stage >= len(self.num_configs)):
			self.finish_up()
			return

		budgets = [self.data[cid].budget for cid in config_ids]
		if len(set(budgets)) > 1:
			raise RuntimeError('Not all configurations have the same budget!')
		budget = self.budgets[self.stage-1]

		losses = np.array([self.data[cid].results[budget]['loss'] for cid in config_ids])

		advance = self._advance_to_next_stage(config_ids, losses)

		for i, a in enumerate(advance):
			if a:
				self.logger.debug('ITERATION: Advancing config %s to next budget %f'%(config_ids[i], self.budgets[self.stage]))

		for i, cid in enumerate(config_ids):
			if advance[i]:
				self.data[cid].status = 'QUEUED'
				self.data[cid].budget = self.budgets[self.stage]
				self.actual_num_configs[self.stage] += 1
			else:
				self.data[cid].status = 'TERMINATED'

	def finish_up(self):
		self.is_finished = True

		for k,v in self.data.items():
			assert v.status in ['TERMINATED', 'REVIEW', 'CRASHED'], 'Configuration has not finshed yet!'
			v.status = 'COMPLETED'


class WarmStartIteration(BaseIteration):
	"""
	iteration that imports a privious Result for warm starting
	"""

	def __init__(self, Result, config_generator):
		
		self.is_finished=False
		self.stage = 0

		
		id2conf = Result.get_id2config_mapping()
		delta_t = - max(map(lambda r: r.time_stamps['finished'], Result.get_all_runs()))

		super().__init__(-1, [len(id2conf)]	, [None], None)
		
		
		for i, id in enumerate(id2conf):
			new_id = self.add_configuration(config=id2conf[id]['config'], config_info=id2conf[id]['config_info'])
			
			for r in Result.get_runs_by_id(id):
			
				
				j = Job(new_id, config=id2conf[id]['config'], budget=r.budget)
				
				j.result = {'loss': r.loss, 'info': r.info}
				j.error_logs = r.error_logs
				
				for k,v in r.time_stamps.items():
					j.timestamps[k] = v + delta_t
				
				self.register_result(j , skip_sanity_checks=True)
				
				config_generator.new_result(j, update_model=(i==len(id2conf)-1))
				
		# mark as finished, as no more runs should be executed from these runs
		self.is_finished = True
		
	def fix_timestamps(self, time_ref):
		"""
			manipulates internal time stamps such that the last run ends at time 0
		"""
		
		for k,v in self.data.items():
			for kk, vv in v.time_stamps.items():
				for kkk,vvv in vv.items():
					self.data[k].time_stamps[kk][kkk] += time_ref
		



import logging
import traceback

class base_config_generator(object):
	"""
	The config generator determines how new configurations are sampled. This can take very different levels of
	complexity, from random sampling to the construction of complex empirical prediction models for promising
	configurations.
	"""
	def __init__(self, logger=None):
		"""
		Parameters
		----------

		directory: string
			where the results are logged
		logger: hpbandster.utils.result_logger_v??
			the logger to store the data, defaults to v1
		overwrite: bool
			whether or not existing data will be overwritten
		logger: logging.logger
			for some debug output

		"""

		if logger is None:
			self.logger=logging.getLogger('hpbandster')
		else:
			self.logger=logger

	def get_config(self, budget):
		"""
		function to sample a new configuration

		This function is called inside Hyperband to query a new configuration

		Parameters
		----------
		budget: float
			the budget for which this configuration is scheduled

		returns: (config, info_dict)
			must return a valid configuration and a (possibly empty) info dict
		"""
		raise NotImplementedError('This function needs to be overwritten in %s.'%(self.__class__.__name__))

	def new_result(self, job, update_model=True):
		"""
		registers finished runs

		Every time a run has finished, this function should be called
		to register it with the result logger. If overwritten, make
		sure to call this method from the base class to ensure proper
		logging.


		Parameters
		----------
		job: instance of hpbandster.distributed.dispatcher.Job
			contains all necessary information about the job
		update_model: boolean
			determines whether a model inside the config_generator should be updated
		"""
		if not job.exception is None:
			self.logger.warning("job {} failed with exception\n{}".format(job.id, job.exception))



import os
import pickle
import json
import threading

import Pyro4.naming


def nic_name_to_host(nic_name):
	""" helper function to translate the name of a network card into a valid host name"""
	from netifaces import ifaddresses, AF_INET
	host = ifaddresses(nic_name).setdefault(AF_INET, [{'addr': 'No IP addr'}] )[0]['addr']
	return(host)


class NameServer(object):
	"""
	The nameserver serves as a phonebook-like lookup table for your workers. Unique names are created so the workers
	can work in parallel and register their results without creating racing conditions. The implementation uses
	`PYRO4 <https://pythonhosted.org/Pyro4/nameserver.html>`_ as a backend and this class is basically a wrapper.
	"""
	def __init__(self, run_id, working_directory=None, host=None, port=0, nic_name=None):
		"""
		Parameters
		----------
			run_id: str
				unique run_id associated with the HPB run
			working_directory: str
				path to the working directory of the HPB run to store the nameservers credentials.
				If None, no config file will be written.
			host: str
				the hostname to use for the nameserver
			port: int
				the port to be used. Default (=0) means a random port
			nic_name: str
				name of the network interface to use (only used if host is not given)
		"""
		self.run_id = run_id
		self.host = host
		self.nic_name = nic_name
		self.port = port
		self.dir = working_directory
		self.conf_fn = None
		self.pyro_ns = None



	def start(self):
		"""	
		starts a Pyro4 nameserver in a separate thread
		
		Returns
		-------
			tuple (str, int):
				the host name and the used port
		"""
	
		if self.host is None:
			if self.nic_name is None:
				self.host = 'localhost'
			else:
				self.host = nic_name_to_host(self.nic_name)

		uri, self.pyro_ns, _ = Pyro4.naming.startNS(host=self.host, port=self.port)

		self.host, self.port = self.pyro_ns.locationStr.split(':')
		self.port = int(self.port)
		
		thread = threading.Thread(target=self.pyro_ns.requestLoop, name='Pyro4 nameserver started by HpBandSter')
		thread.start()

		if not self.dir is None:
			os.makedirs(self.dir, exist_ok=True)
			self.conf_fn = os.path.join(self.dir, 'HPB_run_%s_pyro.pkl'%self.run_id)

			with open(self.conf_fn, 'wb') as fh:
				pickle.dump((self.host, self.port), fh)
		
		return(self.host, self.port)


	def shutdown(self):
		"""
			clean shutdown of the nameserver and the config file (if written)
		"""
		if not self.pyro_ns is None:
			self.pyro_ns.shutdown()
			self.pyro_ns = None
		
		if not self.conf_fn is None:
			os.remove(self.conf_fn)
			self.conf_fn = None


	def __del__(self):
		self.shutdown()






"""
Example 8 - Warmstarting for MNIST
==================================

Sometimes it is desired to continue an already finished run because the optimization
requires more function evaluations. In other cases, one might wish to use results
from previous runs to speed up the optimization. This might be useful if initial
runs were done with relatively small budgets, or on only a subset of the data to
get an initial understanding of the problem.

Here we shall see how to use the results from example 5 to initialize BOHB's model.
What changed are
- the number of training points is increased from 8192 to 32768
- the number of validation points is increased from 1024 to 16384
- the mimum budget is now 3 instead of 1 because we have already quite a few runs for a small number of epochs

Note that the loaded runs will show up in the results of the new run. They are all
combined into an iteration with the index -1 and their time stamps are manipulated
such that the last run finishes at time 0 with all other times being negative.
That info can be used to filter those runs when analysing the run.

"""
import os
import pickle
import argparse

import hpbandster.core.nameserver as hpns
import hpbandster.core.result as hpres

from hpbandster.optimizers import BOHB

import logging
logging.basicConfig(level=logging.DEBUG)



parser = argparse.ArgumentParser(description='Example 1 - sequential and local execution.')
parser.add_argument('--min_budget',   type=float, help='Minimum number of epochs for training.',    default=3)
parser.add_argument('--max_budget',   type=float, help='Maximum number of epochs for training.',    default=9)
parser.add_argument('--n_iterations', type=int,   help='Number of iterations performed by the optimizer', default=4)
parser.add_argument('--worker', help='Flag to turn this into a worker process', action='store_true')
parser.add_argument('--run_id', type=str, help='A unique run id for this optimization run. An easy option is to use the job id of the clusters scheduler.')
parser.add_argument('--nic_name',type=str, help='Which network interface to use for communication.', default='lo')
parser.add_argument('--shared_directory',type=str, help='A directory that is accessible for all processes, e.g. a NFS share.', default='.')
parser.add_argument('--backend',help='Toggles which worker is used. Choose between a pytorch and a keras implementation.', choices=['pytorch', 'keras'], default='keras')
parser.add_argument('--previous_run_dir',type=str, help='A directory that contains a config.json and results.json for the same configuration space.', default='./example_5_run/')

args=parser.parse_args()


if args.backend == 'pytorch':
	from example_5_pytorch_worker import PyTorchWorker as worker
else:
	from example_5_keras_worker import KerasWorker as worker


# Every process has to lookup the hostname
host = hpns.nic_name_to_host(args.nic_name)


if args.worker:
	import time
	time.sleep(5)	# short artificial delay to make sure the nameserver is already running
	w = worker(run_id=args.run_id, host=host, timeout=120)
	w.load_nameserver_credentials(working_directory=args.shared_directory)
	w.run(background=False)
	exit(0)


# This example shows how to log live results. This is most useful
# for really long runs, where intermediate results could already be
# interesting. The core.result submodule contains the functionality to
# read the two generated files (results.json and configs.json) and
# create a Result object.
result_logger = hpres.json_result_logger(directory=args.shared_directory, overwrite=False)


# Start a nameserver:
NS = hpns.NameServer(run_id=args.run_id, host=host, port=0, working_directory=args.shared_directory)
ns_host, ns_port = NS.start()

# Start local worker
w = worker(run_id=args.run_id, host=host, nameserver=ns_host, nameserver_port=ns_port, timeout=120)
w.run(background=True)


# Let us load the old run now to use its results to warmstart a new run with slightly
# different budgets in terms of datapoints and epochs.
# Note that the search space has to be identical though!
previous_run = hpres.logged_results_to_HBS_result(args.previous_run_dir)


# Run an optimizer
bohb = BOHB(  configspace = worker.get_configspace(),
			  run_id = args.run_id,
			  host=host,
			  nameserver=ns_host,
			  nameserver_port=ns_port,
			  result_logger=result_logger,
			  min_budget=args.min_budget, max_budget=args.max_budget, 
			  previous_result = previous_run,				# this is how you tell any optimizer about previous runs
		   )
res = bohb.run(n_iterations=args.n_iterations)

# store results
with open(os.path.join(args.shared_directory, 'results.pkl'), 'wb') as fh:
	pickle.dump(res, fh)

# shutdown
bohb.shutdown(shutdown_workers=True)
NS.shutdown()




from hpbandster.core.base_iteration import BaseIteration
import numpy as np


class SuccessiveHalving(BaseIteration):

	def _advance_to_next_stage(self, config_ids, losses):
		"""
			SuccessiveHalving simply continues the best based on the current loss.
		"""
		ranks = np.argsort(np.argsort(losses))
		return(ranks < self.num_configs[self.stage])



import numpy as np

from hpbandster.learning_curve_models.base import LCModel

from robo.models.lcnet import LCNet, get_lc_net


class LCNetWrapper(LCModel):
    """
        Wrapper around LC-Net
    """

    def __init__(self, max_num_epochs):
        self.max_num_epochs = max_num_epochs
        self.model = LCNet(sampling_method="sghmc",
                           l_rate=np.sqrt(1e-4),
                           mdecay=.05,
                           n_nets=100,
                           burn_in=5000,
                           n_iters=30000,
                           get_net=get_lc_net,
                           precondition=True)
    
    def fit(self, times, losses, configs=None):
        """
            function to train the model on the observed data

            Parameters:
            -----------

            times: list
                list of numpy arrays of the timesteps for each curve
            losses: list
                list of numpy arrays of the loss (the actual learning curve)
            configs: list or None
                list of the configurations for each sample. Each element
                has to be a numpy array. Set to None, if no configuration
                information is available.
        """

        assert np.all(times > 0) and np.all(times <= self.max_num_epochs)

        train = None
        targets = None

        for i in range(len(configs)):

            t_idx = times[i] / self.max_num_epochs

            x = np.repeat(np.array(configs[i])[None, :], t_idx.shape[0], axis=0)
            x = np.concatenate((x, t_idx[:, None]), axis=1)

            # LCNet assumes increasing curves, if we feed in losses here we have to flip the curves
            lc = [1 - l for l in losses[i]]

            if train is None:
                train = x
                targets = lc
            else:
                train = np.concatenate((train, x), 0)
                targets = np.concatenate((targets, lc), 0)

        self.model.train(train, targets)

    def predict_unseen(self, times, config):
        """
            predict the loss of an unseen configuration

            Parameters:
            -----------

            times: numpy array
                times where to predict the loss
            config: numpy array
                the numerical representation of the config

            Returns:
            --------
            
            mean and variance prediction at input times for the given config
        """

        assert np.all(times > 0) and np.all(times <= self.max_num_epochs)

        x = np.array(config)[None, :]

        idx = times / self.max_num_epochs
        x = np.repeat(x, idx.shape[0], axis=0)

        x = np.concatenate((x, idx[:, None]), axis=1)

        mean, var = self.model.predict(x)
        return 1 - mean, var

    def extend_partial(self, times, obs_times, obs_losses, config=None):
        """
            extends a partially observed curve

            Parameters:
            -----------

            times: numpy array
                times where to predict the loss
            obs_times: numpy array
                times where the curve has already been observed
            obs_losses: numpy array
                corresponding observed losses
            config: numpy array
                numerical reperesentation of the config; None if no config
                information is available
                
            Returns:
            --------
            
            mean and variance prediction at input times
                
                
        """
        return self.predict_unseen(times, config)




import sys

from hpbandster.iterations.base import BaseIteration

import numpy as np


class SuccessiveResampling(BaseIteration):
	
	def __init__(self, *args, resampling_rate = 0.5, min_samples_advance = 1, **kwargs):
		"""
			Iteration class to resample new configurations along side keeping the good ones
			in SuccessiveHalving.
			
			Parameters:
			-----------
				resampling_rate: float
					fraction of configurations that are resampled at each stage
				min_samples_advance:int
					number of samples that are guaranteed to proceed to the next
					stage regardless of the fraction.
		
		"""
			self.resampling_rate = resampling_rate
			self.min_samples_advance = min_samples_advance


	def _advance_to_next_stage(self, config_ids, losses):
		"""
			SuccessiveHalving simply continues the best based on the current loss.
		"""
		
		ranks = np.argsort(np.argsort(losses))
		return(ranks < max(self.min_samples_advance, self.num_configs[self.stage] * (1-self.resampling_rate)) )



import os
import threading
import time
import math
import pdb
import copy
import logging

import numpy as np


from hpbandster.core.dispatcher import Dispatcher
from hpbandster.core.result import Result
from hpbandster.core.base_iteration import WarmStartIteration




class Master(object):
	def __init__(self,
			run_id,
			config_generator,
			working_directory='.',
			ping_interval=60,
			nameserver='127.0.0.1',
			nameserver_port=None,
			host=None,
			shutdown_workers=True,
			job_queue_sizes=(-1,0),
			dynamic_queue_size=True,
			logger=None,
			result_logger=None,
			previous_result = None,
			):
		"""The Master class is responsible for the book keeping and to decide what to run next. Optimizers are
                instantiations of Master, that handle the important steps of deciding what configurations to run on what
                budget when.
		
		Parameters
		----------
		run_id : string
			A unique identifier of that Hyperband run. Use, for example, the cluster's JobID when running multiple
			concurrent runs to separate them
		config_generator: hpbandster.config_generators object
			An object that can generate new configurations and registers results of executed runs
		working_directory: string
			The top level working directory accessible to all compute nodes(shared filesystem).
		eta : float
			In each iteration, a complete run of sequential halving is executed. In it,
			after evaluating each configuration on the same subset size, only a fraction of
			1/eta of them 'advances' to the next round.
			Must be greater or equal to 2.
		min_budget : float
			The smallest budget to consider. Needs to be positive!
		max_budget : float
			the largest budget to consider. Needs to be larger than min_budget!
			The budgets will be geometrically distributed :math:`\sim \eta^k` for
			:math:`k\in [0, 1, ... , num\_subsets - 1]`.
		ping_interval: int
			number of seconds between pings to discover new nodes. Default is 60 seconds.
		nameserver: str
			address of the Pyro4 nameserver
		nameserver_port: int
			port of Pyro4 nameserver
		host: str
			ip (or name that resolves to that) of the network interface to use
		shutdown_workers: bool
			flag to control whether the workers are shutdown after the computation is done
		job_queue_size: tuple of ints
			min and max size of the job queue. During the run, when the number of jobs in the queue
			reaches the min value, it will be filled up to the max size. Default: (0,1)
		dynamic_queue_size: bool
			Whether or not to change the queue size based on the number of workers available.
			If true (default), the job_queue_sizes are relative to the current number of workers.
		logger: logging.logger like object
			the logger to output some (more or less meaningful) information
		result_logger: hpbandster.api.results.util.json_result_logger object
			a result logger that writes live results to disk
		previous_result: hpbandster.core.result.Result object
			previous run to warmstart the run
		"""

		self.working_directory = working_directory
		os.makedirs(self.working_directory, exist_ok=True)
		

		if logger is None:
			self.logger = logging.getLogger('hpbandster')
		else:
			self.logger = logger

		self.result_logger = result_logger


		self.config_generator = config_generator
		self.time_ref = None


		self.iterations = []
		self.jobs = []

		self.num_running_jobs = 0
		self.job_queue_sizes = job_queue_sizes
		self.user_job_queue_sizes = job_queue_sizes
		self.dynamic_queue_size = dynamic_queue_size

		if job_queue_sizes[0] >= job_queue_sizes[1]:
			raise ValueError("The queue size range needs to be (min, max) with min<max!")

		if previous_result is None:
			self.warmstart_iteration = []

		else:
			self.warmstart_iteration = [WarmStartIteration(previous_result, self.config_generator)]

		# condition to synchronize the job_callback and the queue
		self.thread_cond = threading.Condition()

		self.config = {
						'time_ref'   : self.time_ref
					}

		self.dispatcher = Dispatcher( self.job_callback, queue_callback=self.adjust_queue_size, run_id=run_id, ping_interval=ping_interval, nameserver=nameserver, nameserver_port=nameserver_port, host=host)

		self.dispatcher_thread = threading.Thread(target=self.dispatcher.run)
		self.dispatcher_thread.start()


	def shutdown(self, shutdown_workers=False):
		self.logger.debug('HBMASTER: shutdown initiated, shutdown_workers = %s'%(str(shutdown_workers)))
		self.dispatcher.shutdown(shutdown_workers)
		self.dispatcher_thread.join()


	def wait_for_workers(self, min_n_workers=1):
		"""
		helper function to hold execution until some workers are active

		Parameters
		----------
		min_n_workers: int
			minimum number of workers present before the run starts		
		"""
	
		self.logger.debug('wait_for_workers trying to get the condition')
		with self.thread_cond:
			while (self.dispatcher.number_of_workers() < min_n_workers):
				self.logger.debug('HBMASTER: only %i worker(s) available, waiting for at least %i.'%(self.dispatcher.number_of_workers(), min_n_workers))
				self.thread_cond.wait(1)
				self.dispatcher.trigger_discover_worker()
				
		self.logger.debug('Enough workers to start this run!')
			

	def get_next_iteration(self, iteration, iteration_kwargs):
		"""
		instantiates the next iteration

		Overwrite this to change the iterations for different optimizers

		Parameters
		----------
			iteration: int
				the index of the iteration to be instantiated
			iteration_kwargs: dict
				additional kwargs for the iteration class

		Returns
		-------
			HB_iteration: a valid HB iteration object
		"""
		
		raise NotImplementedError('implement get_next_iteration for %s'%(type(self).__name__))


	def run(self, n_iterations=1, min_n_workers=1, iteration_kwargs = {},):
		"""
			run n_iterations of SuccessiveHalving

		Parameters
		----------
		n_iterations: int
			number of iterations to be performed in this run
		min_n_workers: int
			minimum number of workers before starting the run
		"""

		self.wait_for_workers(min_n_workers)
		
		iteration_kwargs.update({'result_logger': self.result_logger})

		if self.time_ref is None:
			self.time_ref = time.time()
			self.config['time_ref'] = self.time_ref
		
			self.logger.info('HBMASTER: starting run at %s'%(str(self.time_ref)))

		self.thread_cond.acquire()
		while True:

			self._queue_wait()
			
			next_run = None
			# find a new run to schedule
			for i in self.active_iterations():
				next_run = self.iterations[i].get_next_run()
				if not next_run is None: break

			if not next_run is None:
				self.logger.debug('HBMASTER: schedule new run for iteration %i'%i)
				self._submit_job(*next_run)
				continue
			else:
				if n_iterations > 0:	#we might be able to start the next iteration
					self.iterations.append(self.get_next_iteration(len(self.iterations), iteration_kwargs))
					n_iterations -= 1
					continue

			# at this point there is no imediate run that can be scheduled,
			# so wait for some job to finish if there are active iterations
			if self.active_iterations():
				self.thread_cond.wait()
			else:
				break

		self.thread_cond.release()
		
		for i in self.warmstart_iteration:
			i.fix_timestamps(self.time_ref)
			
		ws_data = [i.data for i in self.warmstart_iteration]
		
		return Result([copy.deepcopy(i.data) for i in self.iterations] + ws_data, self.config)


	def adjust_queue_size(self, number_of_workers=None):

		self.logger.debug('HBMASTER: number of workers changed to %s'%str(number_of_workers))
		with self.thread_cond:
			self.logger.debug('adjust_queue_size: lock accquired')
			if self.dynamic_queue_size:
				nw = self.dispatcher.number_of_workers() if number_of_workers is None else number_of_workers
				self.job_queue_sizes = (self.user_job_queue_sizes[0] + nw, self.user_job_queue_sizes[1] + nw)
				self.logger.info('HBMASTER: adjusted queue size to %s'%str(self.job_queue_sizes))
			self.thread_cond.notify_all()


	def job_callback(self, job):
		"""
		method to be called when a job has finished

		this will do some book keeping and call the user defined
		new_result_callback if one was specified
		"""
		self.logger.debug('job_callback for %s started'%str(job.id))
		with self.thread_cond:
			self.logger.debug('job_callback for %s got condition'%str(job.id))
			self.num_running_jobs -= 1

			if not self.result_logger is None:
				self.result_logger(job)
			self.iterations[job.id[0]].register_result(job)
			self.config_generator.new_result(job)

			if self.num_running_jobs <= self.job_queue_sizes[0]:
				self.logger.debug("HBMASTER: Trying to run another job!")
				self.thread_cond.notify()

		self.logger.debug('job_callback for %s finished'%str(job.id))

	def _queue_wait(self):
		"""
		helper function to wait for the queue to not overflow/underload it
		"""
		
		if self.num_running_jobs >= self.job_queue_sizes[1]:
			while(self.num_running_jobs > self.job_queue_sizes[0]):
				self.logger.debug('HBMASTER: running jobs: %i, queue sizes: %s -> wait'%(self.num_running_jobs, str(self.job_queue_sizes)))
				self.thread_cond.wait()

	def _submit_job(self, config_id, config, budget):
		"""
		hidden function to submit a new job to the dispatcher

		This function handles the actual submission in a
		(hopefully) thread save way
		"""
		self.logger.debug('HBMASTER: trying submitting job %s to dispatcher'%str(config_id))
		with self.thread_cond:
			self.logger.debug('HBMASTER: submitting job %s to dispatcher'%str(config_id))
			self.dispatcher.submit_job(config_id, config=config, budget=budget, working_directory=self.working_directory)
			self.num_running_jobs += 1

		#shouldn't the next line be executed while holding the condition?
		self.logger.debug("HBMASTER: job %s submitted to dispatcher"%str(config_id))

	def active_iterations(self):
		"""
		function to find active (not marked as finished) iterations 

		Returns
		-------
			list: all active iteration objects (empty if there are none)
		"""

		l = list(filter(lambda idx: not self.iterations[idx].is_finished, range(len(self.iterations))))
		return(l)

	def __del__(self):
		pass



import threading
import logging
import queue
import time

import Pyro4


class Job(object):
	def __init__(self, id, **kwargs):
		self.id = id
		
		self.kwargs = kwargs
		
		self.timestamps = {}

		self.result = None
		self.exception = None

		self.worker_name = None

	def time_it(self, which_time):
		self.timestamps[which_time] = time.time()

	def __repr__(self):
		return(\
			"job_id: " +str(self.id) + "\n" + \
			"kwargs: " + str(self.kwargs) + "\n" + \
			"result: " + str(self.result)+ "\n" +\
			"exception: "+ str(self.exception) + "\n"
		)
	def recreate_from_run(self, run):
		
		run.config_id
		run.budget
		run.error_logs  
		run.loss        
		run.info        
		run.time_stamps 




class Worker(object):
	def __init__(self, name, uri):
		self.name = name
		self.proxy = Pyro4.Proxy(uri)
		self.runs_job = None

	def is_alive(self):
		try:
			self.proxy._pyroReconnect(1)
		except Pyro4.errors.ConnectionClosedError:
			return False
		except:
			raise
		return(True)
	
	def shutdown(self):
		self.proxy.shutdown()

	def is_busy(self):
		return(self.proxy.is_busy())
		
	def __repr__(self):
		return(self.name)


class Dispatcher(object):
	"""
	The dispatcher is responsible for assigning tasks to free workers, report results back to the master and
	communicate to the nameserver.
	"""
	def __init__(self, new_result_callback, run_id='0',
					ping_interval=10, nameserver='localhost',
					nameserver_port=None, 
					host=None, logger=None, queue_callback=None):
		"""
		Parameters
		----------
		new_result_callback: function
		    function that will be called with a `Job instance <hpbandster.core.dispatcher.Job>`_ as argument.
		    From the `Job` the result can be read and e.g. logged.
		run_id: str
		    unique run_id associated with the HPB run
		ping_interval: int
		    how often to ping for workers (in seconds)
		nameserver: str
		    address of the Pyro4 nameserver
		nameserver_port: int
		    port of Pyro4 nameserver
		host: str
		    ip (or name that resolves to that) of the network interface to use
		logger: logging.Logger
		    logger-instance for info and debug
		queue_callback: function
		    gets called with the number of workers in the pool on every update-cycle
		"""

		self.new_result_callback = new_result_callback
		self.queue_callback = queue_callback
		self.run_id = run_id
		self.nameserver = nameserver
		self.nameserver_port = nameserver_port
		self.host = host
		self.ping_interval = int(ping_interval)
		self.shutdown_all_threads = False


		if logger is None:
			self.logger = logging.getLogger('hpbandster')
		else:
			self.logger = logger

		self.worker_pool = {}

		self.waiting_jobs = queue.Queue()
		self.running_jobs = {}
		self.idle_workers = set()


		self.thread_lock = threading.Lock()
		self.runner_cond = threading.Condition(self.thread_lock)
		self.discover_cond = threading.Condition(self.thread_lock)

		self.pyro_id="hpbandster.run_%s.dispatcher"%self.run_id


	def run(self):
		with self.discover_cond:
			t1 = threading.Thread(target=self.discover_workers, name='discover_workers')
			t1.start()
			self.logger.info('DISPATCHER: started the \'discover_worker\' thread')
			t2 = threading.Thread(target=self.job_runner, name='job_runner')
			t2.start()
			self.logger.info('DISPATCHER: started the \'job_runner\' thread')
	

			self.pyro_daemon = Pyro4.core.Daemon(host=self.host)

			with Pyro4.locateNS(host=self.nameserver, port=self.nameserver_port) as ns:
				uri = self.pyro_daemon.register(self, self.pyro_id)
				ns.register(self.pyro_id, uri)

			self.logger.info("DISPATCHER: Pyro daemon running on %s"%(self.pyro_daemon.locationStr))

		
		self.pyro_daemon.requestLoop()


		with self.discover_cond:
			self.shutdown_all_threads = True
			self.logger.info('DISPATCHER: Dispatcher shutting down')
			
			self.runner_cond.notify_all()
			self.discover_cond.notify_all()
			
			
		
			with Pyro4.locateNS(self.nameserver, port=self.nameserver_port) as ns:
				ns.remove(self.pyro_id)

		t1.join()
		self.logger.debug('DISPATCHER: \'discover_worker\' thread exited')
		t2.join()
		self.logger.debug('DISPATCHER: \'job_runner\' thread exited')
		self.logger.info('DISPATCHER: shut down complete')

	def shutdown_all_workers(self, rediscover=False):
		with self.discover_cond:
			for worker in self.worker_pool.values():
				worker.shutdown()
			if rediscover:
				time.sleep(1)
				self.discover_cond.notify()

	def shutdown(self, shutdown_workers=False):
		if shutdown_workers:
			self.shutdown_all_workers()

		with self.runner_cond:
			self.pyro_daemon.shutdown()
	
	@Pyro4.expose
	@Pyro4.oneway
	def trigger_discover_worker(self):
		#time.sleep(1)
		self.logger.info("DISPATCHER: A new worker triggered discover_worker")
		with self.discover_cond:
			self.discover_cond.notify()

	
	def discover_workers(self):
		self.discover_cond.acquire()
		sleep_interval = 1
		
		while True:
			self.logger.debug('DISPATCHER: Starting worker discovery')
			update = False
		
			with Pyro4.locateNS(host=self.nameserver, port=self.nameserver_port) as ns:
				worker_names = ns.list(prefix="hpbandster.run_%s.worker."%self.run_id)
				self.logger.debug("DISPATCHER: Found %i potential workers, %i currently in the pool."%(len(worker_names), len(self.worker_pool)))
				
				for wn, uri in worker_names.items():
					if not wn in self.worker_pool:
						w = Worker(wn, uri)
						if not w.is_alive():
							self.logger.debug('DISPATCHER: skipping dead worker, %s'%wn)
							continue 
						update = True
						self.logger.info('DISPATCHER: discovered new worker, %s'%wn)
						self.worker_pool[wn] = w

			# check the current list of workers
			crashed_jobs = set()

			all_workers = list(self.worker_pool.keys())
			for wn in all_workers:
				# remove dead entries from the nameserver
				if not self.worker_pool[wn].is_alive():
					self.logger.info('DISPATCHER: removing dead worker, %s'%wn)
					update = True
					# todo check if there were jobs running on that that need to be rescheduled

					current_job = self.worker_pool[wn].runs_job

					if not current_job is None:
						self.logger.info('Job %s was not completed'%str(current_job))
						crashed_jobs.add(current_job)

					del self.worker_pool[wn]
					self.idle_workers.discard(wn)
					continue
					
				if not self.worker_pool[wn].is_busy():
					self.idle_workers.add(wn)


			# try to submit more jobs if something changed
			if update:
				if not self.queue_callback is None:
					self.discover_cond.release()
					self.queue_callback(len(self.worker_pool))
					self.discover_cond.acquire()
				self.runner_cond.notify()

			for crashed_job in crashed_jobs:
				self.discover_cond.release()
				self.register_result(crashed_job, {'result': None, 'exception': 'Worker died unexpectedly.'})
				self.discover_cond.acquire()

			self.logger.debug('DISPATCHER: Finished worker discovery')

			#if (len(self.worker_pool) == 0 ): # ping for new workers if no workers are currently available
			#	self.logger.debug('No workers available! Keep pinging')
			#	self.discover_cond.wait(sleep_interval)
			#	sleep_interval *= 2
			#else:
			self.discover_cond.wait(self.ping_interval)

			if self.shutdown_all_threads:
				self.logger.debug('DISPATCHER: discover_workers shutting down')
				self.runner_cond.notify()
				self.discover_cond.release()
				return

	def number_of_workers(self):
		with self.discover_cond:
			return(len(self.worker_pool))

	def job_runner(self):
		
		self.runner_cond.acquire()
		while True:
			
			while self.waiting_jobs.empty() or len(self.idle_workers) == 0:
				self.logger.debug('DISPATCHER: jobs to submit = %i, number of idle workers = %i -> waiting!'%(self.waiting_jobs.qsize(),  len(self.idle_workers) ))
				self.runner_cond.wait()
				self.logger.debug('DISPATCHER: Trying to submit another job.')
				if self.shutdown_all_threads:
					self.logger.debug('DISPATCHER: job_runner shutting down')
					self.discover_cond.notify()
					self.runner_cond.release()
					return
			
			job = self.waiting_jobs.get()
			wn = self.idle_workers.pop()

			worker = self.worker_pool[wn]
			self.logger.debug('DISPATCHER: starting job %s on %s'%(str(job.id),worker.name))
		
			job.time_it('started')
			worker.runs_job = job.id
		
			worker.proxy.start_computation(self, job.id, **job.kwargs)

			job.worker_name = wn
			self.running_jobs[job.id] = job

			self.logger.debug('DISPATCHER: job %s dispatched on %s'%(str(job.id),worker.name))


	def submit_job(self, id, **kwargs):
		self.logger.debug('DISPATCHER: trying to submit job %s'%str(id))
		with self.runner_cond:
			job = Job(id, **kwargs)
			job.time_it('submitted')
			self.waiting_jobs.put(job)
			self.logger.debug('DISPATCHER: trying to notify the job_runner thread.')
			self.runner_cond.notify()

	@Pyro4.expose
	@Pyro4.callback
	@Pyro4.oneway
	def register_result(self, id=None, result=None):
		self.logger.debug('DISPATCHER: job %s finished'%(str(id)))
		with self.runner_cond:
			self.logger.debug('DISPATCHER: register_result: lock acquired')
			# fill in missing information
			job = self.running_jobs[id]
			job.time_it('finished')
			job.result = result['result']
			job.exception = result['exception']

			self.logger.debug('DISPATCHER: job %s on %s finished'%(str(job.id),job.worker_name))
			self.logger.debug(str(job))
			
			# delete job
			del self.running_jobs[id]

			# label worker as idle again
			try:
				self.worker_pool[job.worker_name].runs_job = None
				self.worker_pool[job.worker_name].proxy._pyroRelease()
				self.idle_workers.add(job.worker_name)
				# notify the job_runner to check for more jobs to run
				self.runner_cond.notify()
			except KeyError:
				# happens for crashed workers, but we can just continue
				pass
			except:
				raise

		# call users callback function to register the result
		# needs to be with the condition released, as the master can call
		# submit_job quickly enough to cause a dead-lock
		self.new_result_callback(job)



import sys
sys.path.append("../../")

import numpy as np
from sklearn.ensemble import RandomForestRegressor as rfr

from hpbandster.learning_curve_models.base import LCModel as lcm_base

from IPython import embed

class ARIF(lcm_base):
    """
        An 'Auto Regressive Integrated (Random) Forest'
    """
    def __init__(self, order=2, diff_order=0):
        """
            Parameters:
            -----------

            order: int
                the order of the 'autoregressiveness'
            diff_order: int
                the differencing order used
                TODO: Not used so far!
        """
        self.order = order
        self.diff_order = diff_order

    def apply_differencing(self, series, order=None):

        if order is None: order = self.diff_order

        for o in range(order):
            series = series[1:]-series[:-1] 
        return series
            

    def invert_differencing(self, initial_part, differenced_rest, order=None):
        """
            function to invert the differencing
        """

        if order is None: order = self.diff_order

        # compute the differenced values of the initial part:
        starting_points = [ self.apply_differencing(initial_part, order=order)[-1] for order in range(self.diff_order)]
        
        actual_predictions = differenced_rest
        import pdb
        pdb.set_trace()
        for s in starting_points[::-1]:
            actual_predictions = np.cumsum(np.hstack([s, actual_predictions]))[1:]

        return(actual_predictions)

    def fit(self, losses, configs=None):

        if configs is None:
            configs = [[]]*len(times)

        # convert learning curves into X and y data

        X = []
        y = []

        for l,c in zip(losses, configs):
            l = self.apply_differencing(l)

            for i in range(self.order, len(l)):
                X.append(np.hstack([l[i-self.order:i], c]))
                y.append(l[i])

        self.X = np.array(X)
        self.y = np.array(y)


        self.rfr = rfr().fit(self.X,self.y)

        
    def extend_partial(self, obs_losses, num_steps, config=None):
        # TODO: add variance predictions
        if config is None:
            config = []

        d_losses = self.apply_differencing(obs_losses)


        for t in range(num_steps):
            x = np.hstack([d_losses[-self.order:], config])
            y = self.rfr.predict([x])
            d_losses = np.hstack([d_losses, y])


        prediction = self.invert_differencing( obs_losses, d_losses[-num_steps:])

        return(prediction)


if __name__ == "__main__":

    sys.path.append("/home/sfalkner/repositories/bitbucket/learning_curve_prediction")
    

    
    from lc_prediction.utils import load_configs
    
    
    #data = load_configs("/home/sfalkner/repositories/bitbucket/learning_curve_prediction/data/conv_net_cifar10", 256+128)
    data = load_configs("/home/sfalkner/repositories/bitbucket/learning_curve_prediction/data/fc_net_mnist", 1024)
    #data = load_configs("/home/sfalkner/repositories/bitbucket/learning_curve_prediction/data/lr_mnist", 1024)
    
    data = (data[0], data[1][:,:40])
    
    import matplotlib.pyplot as plt
    
    #plt.plot(data[1].T)
    #plt.show()
    
    
    full_lcs =  [ lc for lc in data[1]]
    
    T_max = len(full_lcs[0])
    
    learning_curves = [ lc[:np.random.randint(lc.shape[0]-8) + 8]for lc in data[1]]
    #learning_curves = [ lc[:4+ int(np.random.exponential(5))] for lc in data[1]]
    times = [np.arange(1, lc.shape[0]+1) for lc in learning_curves]
    
    lc_model = ARIF(order=3, diff_order=2)
    
    
    
    test_order = 2
    random_sequence = np.random.rand(5)
    tmp = lc_model.apply_differencing(random_sequence, order=test_order)
    
    for i in range(test_order+1):
        print(lc_model.apply_differencing(random_sequence, order=i))
    reconstruction = lc_model.invert_differencing(random_sequence[:1+test_order], tmp, order=test_order)
    
    embed()
    
    
    
    lc_model.fit(learning_curves, data[0])
    
    for i in range(16):
        pred_times = range(times[i][-1]+1, T_max)
        #pred = lc_model.extend_partial(learning_curves[i], min(10, T_max - len(learning_curves[i])), config=data[0][i])
        pred = lc_model.extend_partial(learning_curves[i], T_max - len(learning_curves[i]), config=data[0][i])
        plt.plot(full_lcs[i])
        plt.plot(range(len(learning_curves[i]), len(learning_curves[i])+ len(pred)), pred, '--')
    plt.show()
    
    embed()
    
    
    



"""
Example 5 - MNIST
=================

Small CNN for MNIST implementet in both Keras and PyTorch.
This example also shows how to log results to disk during the optimization
which is useful for long runs, because intermediate results are directly available
for analysis. It also contains a more realistic search space with different types
of variables to be optimized.

"""
import os
import pickle
import argparse

import hpbandster.core.nameserver as hpns
import hpbandster.core.result as hpres

from hpbandster.optimizers import BOHB

import logging
logging.basicConfig(level=logging.DEBUG)



parser = argparse.ArgumentParser(description='Example 5 - CNN on MNIST')
parser.add_argument('--min_budget',   type=float, help='Minimum number of epochs for training.',    default=1)
parser.add_argument('--max_budget',   type=float, help='Maximum number of epochs for training.',    default=9)
parser.add_argument('--n_iterations', type=int,   help='Number of iterations performed by the optimizer', default=16)
parser.add_argument('--worker', help='Flag to turn this into a worker process', action='store_true')
parser.add_argument('--run_id', type=str, help='A unique run id for this optimization run. An easy option is to use the job id of the clusters scheduler.')
parser.add_argument('--nic_name',type=str, help='Which network interface to use for communication.', default='lo')
parser.add_argument('--shared_directory',type=str, help='A directory that is accessible for all processes, e.g. a NFS share.', default='.')
parser.add_argument('--backend',help='Toggles which worker is used. Choose between a pytorch and a keras implementation.', choices=['pytorch', 'keras'], default='keras')

args=parser.parse_args()


if args.backend == 'pytorch':
	from example_5_pytorch_worker import PyTorchWorker as worker
else:
	from example_5_keras_worker import KerasWorker as worker


# Every process has to lookup the hostname
host = hpns.nic_name_to_host(args.nic_name)


if args.worker:
	import time
	time.sleep(5)	# short artificial delay to make sure the nameserver is already running
	w = worker(run_id=args.run_id, host=host, timeout=120)
	w.load_nameserver_credentials(working_directory=args.shared_directory)
	w.run(background=False)
	exit(0)


# This example shows how to log live results. This is most useful
# for really long runs, where intermediate results could already be
# interesting. The core.result submodule contains the functionality to
# read the two generated files (results.json and configs.json) and
# create a Result object.
result_logger = hpres.json_result_logger(directory=args.shared_directory, overwrite=False)


# Start a nameserver:
NS = hpns.NameServer(run_id=args.run_id, host=host, port=0, working_directory=args.shared_directory)
ns_host, ns_port = NS.start()

# Start local worker
w = worker(run_id=args.run_id, host=host, nameserver=ns_host, nameserver_port=ns_port, timeout=120)
w.run(background=True)

# Run an optimizer
bohb = BOHB(  configspace = worker.get_configspace(),
			  run_id = args.run_id,
			  host=host,
			  nameserver=ns_host,
			  nameserver_port=ns_port,
			  result_logger=result_logger,
			  min_budget=args.min_budget, max_budget=args.max_budget, 
		   )
res = bohb.run(n_iterations=args.n_iterations)

# store results
with open(os.path.join(args.shared_directory, 'results.pkl'), 'wb') as fh:
	pickle.dump(res, fh)

# shutdown
bohb.shutdown(shutdown_workers=True)
NS.shutdown()

